# -*- coding: utf-8 -*-

from ast import literal_eval
from collections import OrderedDict
from collective.contact.plonegroup.config import FUNCTIONS_REGISTRY
from collective.contact.plonegroup.config import ORGANIZATIONS_REGISTRY
from collective.contact.plonegroup.utils import get_all_suffixes
from collective.contact.plonegroup.utils import get_own_organization
from collective.contact.plonegroup.utils import get_plone_group
from collective.contact.plonegroup.utils import get_plone_group_id
from collective.documentgenerator.utils import update_oo_config
from collective.eeafaceted.batchactions.interfaces import IBatchActionsMarker
from copy import deepcopy
from DateTime import DateTime
from datetime import date
from eea.facetednavigation.interfaces import ICriteria
from ftw.labels.interfaces import ILabelJar
from imio.helpers.content import get_vocab_values
from persistent.mapping import PersistentMapping
from plone import api
from plone.app.contenttypes.migration.dxmigration import migrate_base_class_to_new_class
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_unicode
from Products.GenericSetup.tool import DEPENDENCY_STRATEGY_NEW
from Products.PloneMeeting.config import MEETING_GROUP_SUFFIXES
from Products.PloneMeeting.config import TOOL_FOLDER_POD_TEMPLATES
from Products.PloneMeeting.content.person import PMPerson
from Products.PloneMeeting.exportimport.content import update_labels_jar
from Products.PloneMeeting.indexes import DELAYAWARE_ROW_ID_PATTERN
from Products.PloneMeeting.indexes import REAL_ORG_UID_PATTERN
from Products.PloneMeeting.interfaces import IConfigElement
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.PloneMeeting.profiles import MeetingConfigDescriptor
from Products.PloneMeeting.utils import get_public_url
from Products.PloneMeeting.utils import getAdvicePortalTypeIds
from Products.PloneMeeting.utils import updateCollectionCriterion
from Products.PlonePAS.tools.groupdata import GroupData
from z3c.relationfield.relation import RelationValue
from zope.component import getUtility
from zope.i18n import translate
from zope.interface import alsoProvides
from zope.intid.interfaces import IIntIds

import mimetypes
import os


class Migrate_To_4_1(Migrator):

    def _updateFacetedFilters(self):
        """Add new faceted filters :
           For item :
           - 'Has annexes to sign?';
           - 'Labels'.
           For meeting :
           - 'Date'.
           Update vocabulary used for :
           - Creator;
           - Taken over by.
           Update index used for :
           - Group in charge."""
        logger.info("Updating faceted filters for every MeetingConfigs...")

        xmlpath_items = os.path.join(
            os.path.dirname(__file__),
            '../faceted_conf/upgrade_step_add_item_widgets.xml')

        xmlpath_meetings = os.path.join(
            os.path.dirname(__file__),
            '../faceted_conf/upgrade_step_add_meeting_widgets.xml')

        for cfg in self.tool.objectValues('MeetingConfig'):
            obj = cfg.searches.searches_items
            # add new faceted filters for searches_items
            obj.unrestrictedTraverse('@@faceted_exportimport').import_xml(
                import_file=open(xmlpath_items))
            # update vocabulary for relevant filters
            criteria = ICriteria(obj)
            criteria.edit(
                'c4', **{
                    'vocabulary':
                        'Products.PloneMeeting.vocabularies.proposinggroupsforfacetedfiltervocabulary'})
            criteria.edit(
                'c11', **{
                    'vocabulary':
                        'Products.PloneMeeting.vocabularies.creatorsforfacetedfiltervocabulary'})
            criteria.edit(
                'c12', **{
                    'vocabulary':
                        'Products.PloneMeeting.vocabularies.creatorswithnobodyforfacetedfiltervocabulary'})
            criteria.edit(
                'c23', **{
                    'title': 'Groups in charge',
                    'index': 'getGroupsInCharge',
                    'operator_visible': True})

            obj = cfg.searches.searches_meetings
            obj = cfg.searches.searches_decisions
            # add new faceted filters for searches_meetings/searches_items
            obj.unrestrictedTraverse('@@faceted_exportimport').import_xml(
                import_file=open(xmlpath_meetings))
        logger.info('Done.')

    def _addItemTemplatesManagersGroup(self):
        """Add the '_itemtemplatesmanagers' group for every MeetingConfig."""
        logger.info("Adding 'itemtemplatesmanagers' group for every MeetingConfigs...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.createItemTemplateManagersGroup()
        logger.info('Done.')

    def _migrateGroupsInChargeAttributes(self):
        '''Field MeetingConfig.itemGroupInChargeStates was renamed to MeetingConfig.itemGroupsInChargeStates.
           Value reader_groupincharge is now reader_groupsincharge.'''
        logger.info("Updating attributes for groupsInCharge...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if hasattr(cfg, 'itemGroupInChargeStates'):
                itemGroupInChargeStates = cfg.itemGroupInChargeStates
                cfg.setItemGroupsInChargeStates(itemGroupInChargeStates)
                delattr(cfg, 'itemGroupInChargeStates')
            # itemAnnexConfidentialVisibleFor
            itemAnnexConfidentialVisibleFor = list(cfg.getItemAnnexConfidentialVisibleFor())
            if 'reader_groupincharge' in itemAnnexConfidentialVisibleFor:
                itemAnnexConfidentialVisibleFor.remove('reader_groupincharge')
                itemAnnexConfidentialVisibleFor.append('reader_groupsincharge')
            cfg.setItemAnnexConfidentialVisibleFor(itemAnnexConfidentialVisibleFor)
            # adviceAnnexConfidentialVisibleFor
            adviceAnnexConfidentialVisibleFor = list(cfg.getAdviceAnnexConfidentialVisibleFor())
            if 'reader_groupincharge' in adviceAnnexConfidentialVisibleFor:
                adviceAnnexConfidentialVisibleFor.remove('reader_groupincharge')
                adviceAnnexConfidentialVisibleFor.append('reader_groupsincharge')
            cfg.setAdviceAnnexConfidentialVisibleFor(adviceAnnexConfidentialVisibleFor)

        logger.info('Done.')

    def _markSearchesFoldersWithIBatchActionsMarker(self):
        """Mark every searches subfolders with the IBatchActionsMarker."""
        logger.info("Marking members searches folders with the IBatchActionsMarker...")
        for userFolder in self.portal.Members.objectValues():
            mymeetings = getattr(userFolder, 'mymeetings', None)
            if not mymeetings:
                continue
            for cfg in self.tool.objectValues('MeetingConfig'):
                meetingFolder = getattr(mymeetings, cfg.getId(), None)
                if not meetingFolder:
                    continue
                search_folders = [
                    folder for folder in meetingFolder.objectValues('ATFolder')
                    if folder.getId().startswith('searches_')]
                for search_folder in search_folders:
                    if IBatchActionsMarker.providedBy(search_folder):
                        logger.info('Already migrated ...')
                        logger.info('Done.')
                        return
                    alsoProvides(search_folder, IBatchActionsMarker)
        logger.info('Done.')

    def _migratePowerObservers(self):
        """Power observers are now defined in MeetingConfig.powerObservers."""
        logger.info("Migrating power observers to MeetingConfig.powerObservers...")
        default_mc_descr = MeetingConfigDescriptor(None, None, None)
        for cfg in self.tool.objectValues('MeetingConfig'):
            if hasattr(cfg, 'itemPowerObserversStates'):
                power_observers = deepcopy(default_mc_descr.powerObservers)
                for po_infos in power_observers:
                    if po_infos['row_id'] == 'powerobservers':
                        po_infos['item_states'] = cfg.itemPowerObserversStates
                        po_infos['meeting_states'] = cfg.meetingPowerObserversStates
                    elif po_infos['row_id'] == 'restrictedpowerobservers':
                        po_infos['item_states'] = cfg.itemRestrictedPowerObserversStates
                        po_infos['meeting_states'] = cfg.meetingRestrictedPowerObserversStates
                cfg.setPowerObservers(power_observers)
                delattr(cfg, 'itemPowerObserversStates')
                delattr(cfg, 'meetingPowerObserversStates')
                delattr(cfg, 'itemRestrictedPowerObserversStates')
                delattr(cfg, 'meetingRestrictedPowerObserversStates')
                # migrate attributes using powerobservers values
                # hideCssClassesTo
                hideCssClassesTo = cfg.getHideCssClassesTo()
                adapted_hideCssClassesTo = []
                if 'power_observers' in hideCssClassesTo:
                    adapted_hideCssClassesTo.append('powerobservers')
                if 'restricted_power_observers' in hideCssClassesTo:
                    adapted_hideCssClassesTo.append('restrictedpowerobservers')
                cfg.setHideCssClassesTo(adapted_hideCssClassesTo)
                # hideHistoryTo
                hideHistoryTo = cfg.getHideHistoryTo()
                adapted_hideHistoryTo = []
                if 'power_observers' in hideHistoryTo:
                    adapted_hideHistoryTo.append('powerobservers')
                if 'restricted_power_observers' in hideHistoryTo:
                    adapted_hideHistoryTo.append('restrictedpowerobservers')
                cfg.setHideHistoryTo(adapted_hideHistoryTo)
                # hideNotViewableLinkedItemsTo
                hideNotViewableLinkedItemsTo = cfg.getHideNotViewableLinkedItemsTo()
                adapted_hideNotViewableLinkedItemsTo = []
                if 'power_observers' in hideNotViewableLinkedItemsTo:
                    adapted_hideNotViewableLinkedItemsTo.append('powerobservers')
                if 'restricted_power_observers' in hideNotViewableLinkedItemsTo:
                    adapted_hideNotViewableLinkedItemsTo.append('restrictedpowerobservers')
                cfg.setHideNotViewableLinkedItemsTo(adapted_hideNotViewableLinkedItemsTo)
                # adviceConfidentialFor
                adviceConfidentialFor = cfg.getAdviceConfidentialFor()
                adapted_adviceConfidentialFor = []
                if 'power_observers' in adviceConfidentialFor:
                    adapted_adviceConfidentialFor.append('powerobservers')
                if 'restricted_power_observers' in adviceConfidentialFor:
                    adapted_adviceConfidentialFor.append('restrictedpowerobservers')
                cfg.setAdviceConfidentialFor(adapted_adviceConfidentialFor)

        logger.info('Done.')

    def _enableRefusedWFAdaptation(self):
        """State 'refused' is now added by a WF adaptation.
           Check for each MeetingConfig item workflow if it contains a 'refused'
           WF state, if it is the case, enable 'refused' WFAdaptation if available."""
        logger.info("Enabling new WFAdaptation 'refused' if relevant...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            item_wf = self.wfTool.getWorkflowsFor(cfg.getItemTypeName())[0]
            if 'refused' in item_wf.states and 'refused' in get_vocab_values(cfg, 'WorkflowAdaptations'):
                wf_adaptations = list(cfg.getWorkflowAdaptations())
                if 'refused' in wf_adaptations:
                    logger.info('Already migrated ...')
                    logger.info('Done.')
                    return
                wf_adaptations.append('refused')
                cfg.setWorkflowAdaptations(wf_adaptations)
        logger.info('Done.')

    def _removeMCPortalTabs(self):
        """portal_tabs are now generated, remove MC related actions registered
        in portal_actions/portal_tabs."""
        logger.info('Removing MeetingConfig related portal_tabs...')
        actions_to_delete = []
        portal_tabs = self.portal.portal_actions.portal_tabs
        for action_id in portal_tabs:
            if action_id.endswith('_action'):
                actions_to_delete.append(action_id)
        portal_tabs.manage_delObjects(ids=actions_to_delete)
        logger.info('Done.')

    def _manageContentsKeptWhenItemSentToOtherMC(self):
        """Parameter MeetingConfig.keepAdvicesOnSentToOtherMC was replaced by
           MeetingConfig.contentsKeptOnSentToOtherMC."""
        logger.info("Migrating field MeetingConfig.keepAdvicesOnSentToOtherMC to "
                    "MeetingConfig.contentsKeptOnSentToOtherMC...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            if not base_hasattr(cfg, 'keepAdvicesOnSentToOtherMC'):
                self._already_migrated()
                return

            keepAdvicesOnSentToOtherMC = cfg.keepAdvicesOnSentToOtherMC
            contentsKeptOnSentToOtherMC = cfg.getContentsKeptOnSentToOtherMC()
            # we kept advices
            if keepAdvicesOnSentToOtherMC:
                contentsKeptOnSentToOtherMC += ('advices', )
                cfg.setContentsKeptOnSentToOtherMC(contentsKeptOnSentToOtherMC)
            delattr(cfg, 'keepAdvicesOnSentToOtherMC')

        logger.info('Done.')

    def _fixAnnexesMimeType(self):
        """In some cases, mimetype used for annex is not correct because
           it was not found in mimetypes_registry.  Now that we do not use
           mimetypes_registry for this, make sure mimetype used for annexes
           is correct using the mimetypes builtin method."""
        logger.info('Fixing annexes mimetype...')
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog(portal_type=['annex', 'annexDecision'])
        for brain in brains:
            annex = brain.getObject()
            current_content_type = annex.file.contentType
            filename = annex.file.filename
            extension = os.path.splitext(filename)[1].lower()
            mimetype = mimetypes.types_map.get(extension)
            if mimetype and mimetype != current_content_type:
                logger.info('Fixing mimetype for annex at {0}, old was {1}, now will be {2}...'.format(
                    '/'.join(annex.getPhysicalPath()), current_content_type, mimetype))
                annex.file.contentType = mimetype
        logger.info('Done.')

    def _fixPODTemplatesMimeType(self):
        """Mimetype used for POD templates created was 'applicaton/odt',
           we need it to be 'application/vnd.oasis.opendocument.text' so these templates
           are updated by the styles template."""
        logger.info('Fixing POD templates mimetype...')
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog(
            object_provides='collective.documentgenerator.content.pod_template.IConfigurablePODTemplate')
        for brain in brains:
            pod_template = brain.getObject()
            # bypass PodTemplate having a pod_template_to_use
            if pod_template.pod_template_to_use:
                continue
            current_content_type = pod_template.odt_file.contentType
            filename = pod_template.odt_file.filename
            extension = os.path.splitext(filename)[1].lower()
            mimetype = mimetypes.types_map.get(extension)
            if mimetype and mimetype != current_content_type:
                logger.info('Fixing mimetype for POD template at {0}, old was {1}, now will be {2}...'.format(
                    '/'.join(pod_template.getPhysicalPath()), current_content_type, mimetype))
                pod_template.odt_file.contentType = mimetype
        logger.info('Done.')

    def _fixItemsWorkflowHistoryType(self):
        """A bug in ToolPloneMeeting.pasteItems was changing the workflow_history
           to a simple dict.  Make sure existing items use a PersistentMapping."""
        logger.info('Fixing items workflow_history...')
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog(meta_type=['MeetingItem'])
        i = 0
        for brain in brains:
            item = brain.getObject()
            if not isinstance(item.workflow_history, PersistentMapping):
                i = i + 1
                persisted_workflow_history = PersistentMapping(item.workflow_history)
                item.workflow_history = persisted_workflow_history
        logger.info('Fixed workflow_history for {0} items.'.format(i))
        logger.info('Done.')

    def _migrateToDoListSearches(self):
        """Field MeetingConfig.toDoListSearches was a reference field,
           we moved it to a LinesField because new DashboardCollection
           are not referenceable by default."""
        logger.info('Migrating to do searches...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            # get references from at_references so order is kept
            reference_uids = [ref.targetUID for ref in cfg.at_references.objectValues()
                              if ref.relationship == 'ToDoSearches']
            if reference_uids:
                cfg.deleteReferences('ToDoSearches')
                cfg.setToDoListSearches(reference_uids)
        logger.info('Done.')

    def _upgradeImioDashboard(self):
        """Move to eea.facetednavigation 10+."""
        logger.info('Upgrading imio.dashboard...')
        catalog = self.portal.portal_catalog
        # before upgrading profile, we must save DashboardCollection that are not enabled
        # before we used a workflow state but now it is a attribute 'enabled' on the DashboardCollection
        brains = catalog(portal_type='DashboardCollection', review_state='inactive')
        disabled_collection_uids = [brain.UID for brain in brains]
        self.upgradeProfile('imio.dashboard:default')
        # now disable relevant DashboardCollections
        brains = catalog(UID=disabled_collection_uids)
        for brain in brains:
            collection = brain.getObject()
            collection.enabled = False
            collection.reindexObject()
        # need to adapt fields maxShownListings, maxShownListings and maxShownAvailableItems
        # of MeetingConfig that are now integers
        for cfg in self.tool.objectValues('MeetingConfig'):
            # maxShownListings
            field = cfg.getField('maxShownListings')
            value = field.get(cfg)
            field.set(cfg, int(value))
            # maxShownAvailableItems
            field = cfg.getField('maxShownAvailableItems')
            value = field.get(cfg)
            field.set(cfg, int(value))
            # maxShownMeetingItems
            field = cfg.getField('maxShownMeetingItems')
            value = field.get(cfg)
            field.set(cfg, int(value))
            # remove old topics folder and thus contained Topics
            if 'topics' in cfg.objectIds():
                api.content.delete(obj=cfg.topics)
        # old forget, set global_allow to False on Topic
        topic = self.portal.portal_types.get('Topic')
        if topic:
            topic.global_allow = False
        logger.info('Done.')

    def _adaptForContacts(self):
        """Add new attributes 'orderedContacts' and 'itemAbsents/itemExcused'
           to existing meetings.
           Remove attribute 'itemAbsents' from existing items."""
        logger.info('Adapting application for contacts...')
        contacts = self.portal.contacts
        own_org = get_own_organization()
        catalog = api.portal.get_tool('portal_catalog')
        intids = getUtility(IIntIds)
        for cfg in self.tool.objectValues('MeetingConfig'):
            logger.info('Migrating config {0}...'.format(cfg.getId()))
            # give local_roles to relevant groups on contacts
            cfg._createOrUpdateAllPloneGroups(force_update_access=True)
            if not hasattr(cfg, 'useUserReplacements'):
                self._already_migrated(False)
                break
            delattr(cfg, 'useUserReplacements')
            # first migrate MeetingUsers
            logger.info('Migrating MeetingUsers...')
            mu_hp_mappings = {}
            wfTool = api.portal.get_tool('portal_workflow')
            for mu in cfg.meetingusers.objectValues('MeetingUser'):
                gender = mu.getGender().upper()
                if gender == u'M':
                    person_title = u'Mr'
                elif gender == u'F':
                    person_title = u'Mrs'
                else:
                    person_title = u'-'
                splitted_name = mu.title.split(u' ')
                person_data = {
                    'id': mu.getId(),
                    'lastname': splitted_name[-1],
                    'firstname': u' '.join(splitted_name[0:-1]).strip(),
                    'gender': gender,
                    'person_title': translate(person_title,
                                              domain='collective.contact.core',
                                              context=self.request)
                }
                hp_data = {
                    'id': mu.getId() + '_hp1',
                    'position': RelationValue(intids.getId(own_org)),
                    'label': unicode(mu.getDuty(), 'utf-8'),
                    'usages': [usage for usage in mu.getUsages() if usage in ['assemblyMember', 'asker']],
                    'signature_number': 'signer' in mu.getUsages() and '1' or None}
                # person may already exist as MeetingUsers were created on a per MeetingConfig basis
                if person_data['id'] not in contacts.objectIds():
                    person = api.content.create(container=contacts, type='person', **person_data)
                    hp = api.content.create(container=person, type='held_position', **hp_data)
                else:
                    hp = contacts.get(person_data['id']).get(hp_data['id'])

                if wfTool.getInfoFor(mu, 'review_state') == 'inactive':
                    self.portal.portal_workflow.doActionFor(hp, 'deactivate')
                mu_hp_mappings[mu.getId()] = hp.UID()

            # migrate Meetings
            brains = catalog(portal_type=cfg.getMeetingTypeName())
            logger.info('Adapting meetings...')
            for brain in brains:
                meeting = brain.getObject()
                if hasattr(meeting, 'orderedContacts'):
                    self._already_migrated(False)
                    break
                meeting.orderedContacts = OrderedDict()
                meeting.itemAbsents = PersistentMapping()
                meeting.itemExcused = PersistentMapping()
                meeting.itemNonAttendees = PersistentMapping()
                meeting.itemSignatories = PersistentMapping()
                # migrate MeetingUsers related fields
                if not meeting.getAssembly():
                    # build attendees and signatories passed to Meeting._doUpdateContacts
                    # attendees OrderedDict([('uid1', 'attendee'), ('uid2', 'attendee'), ('uid3', 'absent')])
                    # signatories {'uid1': '1'}
                    attendees = OrderedDict()
                    signatories = {}
                    for attendee_id in meeting.attendees:
                        hp_uid = mu_hp_mappings.get(attendee_id)
                        if hp_uid:
                            attendees[hp_uid] = 'attendee'
                        else:
                            self.warn(
                                logger,
                                'Could not find attendee "{0}" selected on meeting at "{1}"'.format(
                                    attendee_id, '/'.join(meeting.getPhysicalPath())))
                    for absent_id in meeting.absents:
                        hp_uid = mu_hp_mappings.get(absent_id)
                        if hp_uid:
                            attendees[hp_uid] = 'absent'
                        else:
                            self.warn(
                                logger,
                                'Could not find absent "{0}" selected on meeting at "{1}"'.format(
                                    attendee_id, '/'.join(meeting.getPhysicalPath())))
                    for excused_id in meeting.excused:
                        hp_uid = mu_hp_mappings.get(excused_id)
                        if hp_uid:
                            attendees[hp_uid] = 'excused'
                        else:
                            self.warn(
                                logger,
                                'Could not find excused "{0}" selected on meeting at "{1}"'.format(
                                    attendee_id, '/'.join(meeting.getPhysicalPath())))
                    if not meeting.getSignatures():
                        signature_number = 1
                        for signatory_id in meeting.signatories:
                            hp_uid = mu_hp_mappings.get(signatory_id)
                            if hp_uid:
                                signatories[hp_uid] = str(signature_number)
                                signature_number += 1
                            else:
                                self.warn(
                                    logger,
                                    'Could not find signatory "{0}" selected on meeting at "{1}"'.format(
                                        signatory_id, '/'.join(meeting.getPhysicalPath())))
                    meeting._doUpdateContacts(attendees=attendees, signatories=signatories)

                # remove old informations about MeetingUsers
                delattr(meeting, 'attendees')
                delattr(meeting, 'lateAttendees')
                delattr(meeting, 'absents')
                delattr(meeting, 'excused')
                delattr(meeting, 'signatories')

            # migrate MeetingItems
            logger.info('Adapting items...')
            brains = catalog(portal_type=cfg.getItemTypeName())
            for brain in brains:
                item = brain.getObject()
                if not hasattr(item, 'itemAbsents'):
                    self._already_migrated(False)
                    break
                # migrate MeetingUsers related fields
                itemInitiators = item.getItemInitiator()
                if itemInitiators:
                    # do not fail if a MeetingUser was removed
                    item.setItemInitiator([mu_hp_mappings[itemInitiator] for itemInitiator in itemInitiators
                                           if itemInitiator in mu_hp_mappings])
                    for itemInitiator in itemInitiators:
                        if itemInitiator not in mu_hp_mappings:
                            self.warn(
                                logger,
                                'Could not find itemInitiator "{0}" selected on item at "{1}"'.format(
                                    itemInitiator, '/'.join(item.getPhysicalPath())))
                delattr(item, 'itemAbsents')
                delattr(item, 'itemSignatories')
                delattr(item, 'answerers')
                delattr(item, 'questioners')
                item.votes = PersistentMapping()

            # set orderedContacts and orderedItemInitiators
            # then remove the meetingusers folder and contained MeetingUsers
            orderedContacts = [
                mu_hp_mappings[mu.id] for mu in cfg.meetingusers.objectValues('MeetingUser')
                if mu.queryState() == 'active' and 'assemblyMember' in mu.getUsages()]
            cfg.setOrderedContacts(orderedContacts)
            orderedItemInitiators = [
                mu_hp_mappings[mu.id] for mu in cfg.meetingusers.objectValues('MeetingUser')
                if mu.queryState() == 'active' and 'asker' in mu.getUsages()]
            cfg.setOrderedItemInitiators(orderedItemInitiators)

            cfg.manage_delObjects(ids=['meetingusers'])

        logger.info('Done.')

    def _custom_migrate_meeting_group_to_org(self, mGroup, org):
        """Hook for plugins that need to migrate extra data from MeetingGroup to organization."""
        pass

    def _hook_before_mgroups_to_orgs(self):
        """Hook for plugins that need to do things just before MeetingGroup to organization migration."""
        pass

    def _hook_after_mgroups_to_orgs(self):
        """Hook for plugins that need to do things just after MeetingGroup to organization migration,
           when MeetingGroups and organizations still coexist."""
        pass

    def _adaptForPlonegroup(self):
        """Migrate MeetingGroups to contacts and configure plonegroup.
           Migrate also every relations to the organization as we used the id and we use now the uid."""
        logger.info('Adapting application for plonegroup...')
        own_org = get_own_organization()
        if own_org.objectValues():
            self._already_migrated()
            return

        logger.info('Migrating MeetingGroups...')

        # call hook for external profiles to migrate their MeetingGroups customizations if necessary
        self._hook_before_mgroups_to_orgs()

        enabled_orgs = []
        every_orgs = []
        for mGroup in self.tool.objectValues('MeetingGroup'):
            cs = mGroup.getCertifiedSignatures()
            # empty dates must be None, signatureNumber is now signature_number
            # dates were stored as str, we need datetime now
            adapted_cs = [
                {'signature_number': sign['signatureNumber'],
                 'name': safe_unicode(sign['name']),
                 'function': safe_unicode(sign['function']),
                 'held_position': None,
                 'date_from': sign['date_from'] and date.fromtimestamp(int(DateTime(sign['date_from']))) or None,
                 'date_to': sign['date_to'] and date.fromtimestamp(int(DateTime(sign['date_to']))) or None}
                for sign in cs]
            data = {'id': mGroup.getId(),
                    'title': mGroup.Title(),
                    'description': mGroup.Description(),
                    'acronym': mGroup.getAcronym(),
                    'item_advice_states': mGroup.getItemAdviceStates(),
                    'item_advice_edit_states': mGroup.getItemAdviceEditStates(),
                    'item_advice_view_states': mGroup.getItemAdviceViewStates(),
                    'keep_access_to_item_when_advice_is_given': mGroup.getKeepAccessToItemWhenAdviceIsGiven(),
                    'as_copy_group_on': mGroup.getAsCopyGroupOn(),
                    'certified_signatures': adapted_cs,
                    'groups_in_charge': mGroup.getGroupsInCharge(), }
            new_org = api.content.create(container=own_org, type='organization', **data)
            new_org_uid = new_org.UID()
            if mGroup.queryState() == 'active':
                enabled_orgs.append(new_org_uid)
            every_orgs.append(new_org_uid)
            self._custom_migrate_meeting_group_to_org(mGroup, new_org)

        # configure Plonegroup
        functions = deepcopy(MEETING_GROUP_SUFFIXES)
        # extra suffixes
        from Products.PloneMeeting.config import EXTRA_GROUP_SUFFIXES
        functions = functions + deepcopy(EXTRA_GROUP_SUFFIXES)
        # now replace group_id in 'fct_orgs' by corresponding org uid and translate fct_title
        adapted_functions = []
        for function in functions:
            adapted_function = {'fct_id': function['fct_id'],
                                'fct_title': translate(
                                    function['fct_title'],
                                    domain='PloneMeeting',
                                    context=self.request)}
            adapted_function_orgs = [own_org.get(group_id).UID() for group_id in function['fct_orgs']]
            adapted_function['fct_orgs'] = adapted_function_orgs
            adapted_function['enabled'] = function['enabled']
            adapted_functions.append(adapted_function)
        api.portal.set_registry_record(FUNCTIONS_REGISTRY, adapted_functions)
        # first set every organizations so every subgroups are created
        # then set only enabled orgs
        api.portal.set_registry_record(ORGANIZATIONS_REGISTRY, every_orgs)
        api.portal.set_registry_record(ORGANIZATIONS_REGISTRY, enabled_orgs)

        logger.info('Transfering users to new Plone groups...')
        # transfer users to new Plone groups
        portal_groups = api.portal.get_tool('portal_groups')
        ori_plone_group_ids_to_remove = []
        mc_ids = tuple([mc.getId() for mc in self.tool.objectValues('MeetingConfig')])
        for mGroup in self.tool.objectValues('MeetingGroup'):
            org = own_org.get(mGroup.getId())
            org_uid = org.UID()
            for suffix in get_all_suffixes(org_uid):
                ori_plone_group_id = mGroup.getPloneGroupId(suffix)
                ori_plone_group = api.group.get(ori_plone_group_id)
                if ori_plone_group:
                    new_plone_group = get_plone_group(org_uid, suffix)
                    # manage also rare cases where groups stored on groups
                    for member_or_group in ori_plone_group.getGroupMembers():
                        # no need to manage unexisting as getGroupMembers
                        # only return found users/groups
                        # if member_or_group is a group, we need to get new group based on org_uid
                        if isinstance(member_or_group, GroupData):
                            ori_plone_subgroup_id, subgroup_suffix = member_or_group.getId().rsplit('_', 1)
                            suborg = own_org.get(ori_plone_subgroup_id)
                            if suborg:
                                suborg_uid = suborg.UID()
                                new_plone_subgroup = get_plone_group(suborg_uid, subgroup_suffix)
                                member_or_group = new_plone_subgroup
                            else:
                                # a group like 'meeting-config-id_meetingmanagers'
                                if not member_or_group.getId().startswith(mc_ids):
                                    self.warn(logger, 'Could not add group {0} to {1}'.format(
                                        member_or_group.getId(), new_plone_group.getId()))
                                    continue
                        api.group.add_user(group=new_plone_group, user=member_or_group)
                    # do not remove immediatelly as a group can be stored in another group
                    ori_plone_group_ids_to_remove.append(ori_plone_group_id)
        for ori_plone_group_id in ori_plone_group_ids_to_remove:
            portal_groups.removeGroup(ori_plone_group_id)

        # now that every groups are migrated, we may migrate groups_in_charge
        # we have old MeetingGroup ids stored, we want organization UIDs
        for org in own_org.objectValues():
            if org.groups_in_charge:
                groups_in_charge = [own_org.get(gic_id).UID() for gic_id in org.groups_in_charge]
                org.groups_in_charge = groups_in_charge

        logger.info('Migrating MeetingConfigs...')
        # adapt MeetingConfigs
        for cfg in self.tool.objectValues('MeetingConfig'):
            # migrate DashboardCollections using indexAdvisers
            for brain in api.content.find(context=cfg.searches, portal_type='DashboardCollection'):
                dc = brain.getObject()
                query = dc.query
                found = False
                adapted_query = []
                for line in query:
                    adapted_line = line.copy()
                    if line['i'] == 'indexAdvisers':
                        found = True
                        new_v = []
                        for old_v in line['v']:
                            if old_v.startswith('real_group_id__'):
                                # it is a group_id, turn it to org_uid
                                prefix, group_id = old_v.split('real_group_id__')
                                new_v.append(REAL_ORG_UID_PATTERN.format(own_org.get(group_id).UID()))
                            elif old_v.startswith('delay_real_group_id__'):
                                # row_id, just change prefix
                                new_v.append(old_v.replace(
                                    'delay_real_group_id__',
                                    DELAYAWARE_ROW_ID_PATTERN.format('')))
                        adapted_line['v'] = new_v
                    adapted_query.append(adapted_line)
                if found:
                    dc.query = adapted_query
                    dc._p_changed = True

            # migrate MeetingCategories using usingGroups
            categories = cfg.categories.objectValues('MeetingCategory')
            classifiers = cfg.classifiers.objectValues('MeetingCategory')
            for cat in tuple(categories) + tuple(classifiers):
                usingGroups = cat.getUsingGroups()
                if usingGroups:
                    adapted_usingGroups = []
                    for mGroupId in usingGroups:
                        org = own_org.get(mGroupId)
                        adapted_usingGroups.append(org.UID())
                    cat.setUsingGroups(adapted_usingGroups)

            # advicesKeptOnSentToOtherMC
            advicesKeptOnSentToOtherMC = cfg.getAdvicesKeptOnSentToOtherMC()
            adapted_advicesKeptOnSentToOtherMC = []
            for old_v in advicesKeptOnSentToOtherMC:
                new_value = old_v
                if old_v.startswith('real_group_id__'):
                    prefix, group_id = old_v.split('real_group_id__')
                    new_value = REAL_ORG_UID_PATTERN.format(own_org.get(group_id).UID())
                else:
                    new_value = old_v.replace('delay_real_group_id__', DELAYAWARE_ROW_ID_PATTERN.format(''))
                adapted_advicesKeptOnSentToOtherMC.append(new_value)
            cfg.setAdvicesKeptOnSentToOtherMC(adapted_advicesKeptOnSentToOtherMC)
            # certifiedSignatures
            certifiedSignatures = cfg.getCertifiedSignatures()
            adapted_certifiedSignatures = []
            for v in certifiedSignatures:
                new_value = v.copy()
                new_value['held_position'] = '_none_'
                adapted_certifiedSignatures.append(new_value)
            cfg.setCertifiedSignatures(adapted_certifiedSignatures)
            # customAdvisers
            customAdvisers = cfg.getCustomAdvisers()
            adapted_customAdvisers = []
            for v in customAdvisers:
                new_value = v.copy()
                mGroupId = new_value.pop('group')
                org = own_org.get(mGroupId)
                new_value['org'] = org.UID()
                adapted_customAdvisers.append(new_value)
            cfg.setCustomAdvisers(adapted_customAdvisers)
            # groupsHiddenInDashboardFilter
            groupsHiddenInDashboardFilter = cfg.getGroupsHiddenInDashboardFilter()
            adapted_groupsHiddenInDashboardFilter = []
            for v in groupsHiddenInDashboardFilter:
                org = own_org.get(v)
                adapted_groupsHiddenInDashboardFilter.append(org.UID())
            cfg.setGroupsHiddenInDashboardFilter(adapted_groupsHiddenInDashboardFilter)
            # powerAdvisersGroups
            powerAdvisersGroups = cfg.getPowerAdvisersGroups()
            adapted_powerAdvisersGroups = []
            for v in powerAdvisersGroups:
                org = own_org.get(v)
                adapted_powerAdvisersGroups.append(org.UID())
            cfg.setPowerAdvisersGroups(adapted_powerAdvisersGroups)
            # selectableAdvisers
            selectableAdvisers = cfg.getSelectableAdvisers()
            adapted_selectableAdvisers = []
            for v in selectableAdvisers:
                org = own_org.get(v)
                adapted_selectableAdvisers.append(org.UID())
            cfg.setSelectableAdvisers(adapted_selectableAdvisers)
            # selectableCopyGroups
            selectableCopyGroups = cfg.getSelectableCopyGroups()
            adapted_selectableCopyGroups = []
            for v in selectableCopyGroups:
                mGroupId, suffix = v.rsplit('_', 1)
                org = own_org.get(mGroupId)
                new_value = get_plone_group_id(org.UID(), suffix)
                adapted_selectableCopyGroups.append(new_value)
            cfg.setSelectableCopyGroups(adapted_selectableCopyGroups)

        # adapt MeetingItems
        brains = api.content.find(meta_type='MeetingItem')
        len_brains = len(brains)
        logger.info('Migrating {0} MeetingItems...'.format(len_brains))
        portal_repository = api.portal.get_tool('portal_repository')
        portal_historiesstorage = api.portal.get_tool('portal_historiesstorage')
        histories_repo = portal_historiesstorage._getZVCRepo()
        i = 1
        for brain in api.content.find(meta_type='MeetingItem'):
            item = brain.getObject()
            logger.info('Migrating MeetingItem contacts {0}/{1} ({2})'.format(
                i, len_brains, '/'.join(item.getPhysicalPath())))
            i = i + 1
            # adviceIndex
            adviceIndex_copy = item.adviceIndex.copy()
            for mGroupId in adviceIndex_copy:
                org = own_org.get(mGroupId)
                org_uid = org.UID()
                value = adviceIndex_copy[mGroupId].copy()
                value['id'] = org_uid
                item.adviceIndex.pop(mGroupId)
                item.adviceIndex[org_uid] = value
                item.adviceIndex._p_changed = True
            # copyGroups (autoCopyGroups are updated automatically)
            copyGroups = item.getCopyGroups()
            adapted_copyGroups = []
            for v in copyGroups:
                mGroupId, suffix = v.rsplit('_', 1)
                realGroupId = item._realCopyGroupId(mGroupId)
                org = own_org.get(realGroupId)
                if not org:
                    self.warn(logger, 'Ignoring value "{0}" when migrating copyGroups on item at "{1}"'.format(
                        v, '/'.join(item.getPhysicalPath())))
                    continue
                new_value = get_plone_group_id(org.UID(), suffix)
                adapted_copyGroups.append(new_value)
            item.setCopyGroups(adapted_copyGroups)
            # proposingGroup
            proposingGroup = item.getProposingGroup()
            if proposingGroup:
                proposing_group_org_uid = own_org.get(proposingGroup).UID()
                item.setProposingGroup(proposing_group_org_uid)
            # groupInCharge is moved to groupsInCharge
            groupInCharge = item.groupInCharge
            if groupInCharge:
                group_in_charge = own_org.get(groupInCharge)
                item.setProposingGroupWithGroupInCharge(
                    '{0}__groupincharge__{1}'.format(
                        proposing_group_org_uid, group_in_charge.UID()))
            else:
                item.getField('proposingGroupWithGroupInCharge').set(item, '')
            delattr(item, 'groupInCharge')
            # optionalAdvisers
            optionalAdvisers = item.getOptionalAdvisers()
            adapted_optionalAdvisers = []
            for mGroupId in optionalAdvisers:
                realGroupId = mGroupId.split('__rowid__')[0]
                org = own_org.get(realGroupId)
                new_value = mGroupId.replace(realGroupId, org.UID())
                adapted_optionalAdvisers.append(new_value)
            item.setOptionalAdvisers(adapted_optionalAdvisers)
            # associatedGroups
            associatedGroups = item.getAssociatedGroups()
            if associatedGroups:
                associated_group_org_uids = [own_org.get(associatedGroup).UID()
                                             for associatedGroup in associatedGroups]
                item.setAssociatedGroups(associated_group_org_uids)
            # templateUsingGroups
            templateUsingGroups = item.getTemplateUsingGroups()
            if templateUsingGroups:
                adapted_templateUsingGroups = []
                for mGroupId in templateUsingGroups:
                    org = own_org.get(mGroupId)
                    # templateUsingGroup could contains removed MeetingGroups
                    if org:
                        adapted_templateUsingGroups.append(org.UID())
                item.setTemplateUsingGroups(adapted_templateUsingGroups)

            # adapt contained advices
            for advice in item.getAdvices():
                old_advice_group = advice.advice_group
                advice_org = own_org.get(old_advice_group)
                advice_org_uid = advice_org.UID()
                advice.advice_group = advice_org_uid
                # remove old local_role given to 'Editor' MeetingGroup id
                # we do not add it back as it is managed by dexterity.localrolesfield now
                advice.manage_delLocalRoles(['{0}_advisers'.format(old_advice_group)])

                # migrate versionned advices...
                # we need to get the stored data because using CMFEditions
                # api will always return wrapped data or copy
                history_metadata = portal_repository.getHistoryMetadata(advice)
                if history_metadata:
                    for version_info in history_metadata._full.values():
                        vc_info = version_info['vc_info']
                        version_advice = histories_repo._histories[vc_info.history_id]._versions[
                            vc_info.version_id]._data._object.object
                        if version_advice:
                            org = own_org.get(version_advice.advice_group)
                            if org:
                                version_advice.advice_group = org.UID()
                            else:
                                logger.info('Could not find advice_group {0} for advice version of {1}!'.format(
                                    version_advice.advice_group, advice.absolute_url_path()))
                        else:
                            logger.info('Could not find version_id {0} for advice version of {1}!'.format(
                                vc_info.version_id, advice.absolute_url_path()))

        # MeetingGroups and organizations coexist, call hook
        self._hook_after_mgroups_to_orgs()

        # update every items local roles when every items have been updated because
        # linked items (predecessor) may be updated during this process and we have
        # to make sure their values were already updated
        # + update local roles will also fix 'delay_when_stopped' on advice with delay
        self.tool.updateAllLocalRoles(meta_type=('MeetingItem', ))

        # remove MeetingGroup objects and portal_type
        m_group_ids = [mGroup.getId() for mGroup in self.tool.objectValues('MeetingGroup')]
        self.tool.manage_delObjects(ids=m_group_ids)
        logger.info('Done.')

    def _updateUsedAttributes(self):
        """Now that 'MeetingItem.description' is an optional field, we need to
           select it on existing MeetingConfigs.
           Remove 'MeetingItem.itemAssembly' from selected values as it is no more an
           optional field and is used if 'Meeting.assembly' is used.
           Remove also fields removed from MeetingItem schema."""
        logger.info('Updating every MeetingConfig.usedItemAttributes/MeetingConfig.usedMeetingAttributes...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            # MeetingItem
            usedItemAttrs = list(cfg.getUsedItemAttributes())
            if 'description' not in usedItemAttrs:
                usedItemAttrs.insert(0, 'description')
            if 'itemAssembly' in usedItemAttrs:
                usedItemAttrs.remove('itemAssembly')
            if 'questioners' in usedItemAttrs:
                usedItemAttrs.remove('questioners')
            if 'answerers' in usedItemAttrs:
                usedItemAttrs.remove('answerers')
            if 'lateAttendees' in usedItemAttrs:
                usedItemAttrs.remove('lateAttendees')
            cfg.setUsedItemAttributes(usedItemAttrs)
            # Meeting
            usedMeetingAttrs = list(cfg.getUsedMeetingAttributes())
            if 'lateAttendees' in usedMeetingAttrs:
                usedMeetingAttrs.remove('lateAttendees')
            cfg.setUsedMeetingAttributes(usedMeetingAttrs)
        logger.info('Done.')

    def _updateHistorizedAttributes(self):
        """Some fields were removed from MeetingItem/Meeting schema."""
        logger.info(
            'Updating every MeetingConfig.historizedItemAttributes/MeetingConfig.historizedMeetingAttributes...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            # MeetingItem
            histItemAttrs = list(cfg.getHistorizedItemAttributes())
            if 'questioners' in histItemAttrs:
                histItemAttrs.remove('questioners')
            if 'answerers' in histItemAttrs:
                histItemAttrs.remove('answerers')
            if 'itemSignatories' in histItemAttrs:
                histItemAttrs.remove('itemSignatories')
            if 'itemAbsents' in histItemAttrs:
                histItemAttrs.remove('itemAbsents')
            cfg.setHistorizedItemAttributes(histItemAttrs)
            # Meeting
            histMeetingAttrs = list(cfg.getHistorizedMeetingAttributes())
            if 'attendees' in histMeetingAttrs:
                histMeetingAttrs.remove('attendees')
            if 'excused' in histMeetingAttrs:
                histMeetingAttrs.remove('excused')
            if 'absents' in histMeetingAttrs:
                histMeetingAttrs.remove('absents')
            if 'lateAttendees' in histMeetingAttrs:
                histMeetingAttrs.remove('lateAttendees')
            cfg.setHistorizedMeetingAttributes(histMeetingAttrs)
        logger.info('Done.')

    def _migrateGroupsShownInDashboardFilter(self):
        """MeetingConfig.groupsHiddenInDashboardFilter was MeetingConfig.groupsShownInDashboardFilter."""
        logger.info('Migrating "MeetingConfig.groupsShownInDashboardFilter" to '
                    '"MeetingConfig.groupsHiddenInDashboardFilter"...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            if not hasattr(cfg, 'groupsShownInDashboardFilter'):
                self._already_migrated(False)
                break
            # if groupsShownInDashboardFilter was empty, it meants, we show every groups
            groupsShownInDashboardFilter = cfg.groupsShownInDashboardFilter
            if groupsShownInDashboardFilter:
                own_org = get_own_organization()
                adapted_groupsShownInDashboardFilter = [own_org.get(group_id).UID() for group_id
                                                        in groupsShownInDashboardFilter]
                org_uids = cfg.getField('groupsHiddenInDashboardFilter').Vocabulary(cfg).keys()
                new_values = [org_uid for org_uid in org_uids
                              if org_uid not in adapted_groupsShownInDashboardFilter]
                cfg.setGroupsHiddenInDashboardFilter(new_values)
            delattr(cfg, 'groupsShownInDashboardFilter')
        logger.info('Done.')

    def _enableStyleTemplates(self):
        """Content type StyleTemplate is now added to meetingConfigs."""
        logger.info("Enabling StyleTemplate ...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            folder = cfg.get(TOOL_FOLDER_POD_TEMPLATES)
            allowed_content_types = folder.getLocallyAllowedTypes()
            if 'StyleTemplate' not in allowed_content_types:
                allowed_content_types += ('StyleTemplate',)
                folder.setLocallyAllowedTypes(allowed_content_types)
                folder.reindexObject()
        logger.info('Done.')

    def _fixMeetingCollectionsQuery(self):
        """The review_state value must be a list, not a tuple."""
        logger.info('Fixing meetings related collections query...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            collections = tuple(cfg.searches.searches_meetings.objectValues()) + \
                tuple(cfg.searches.searches_decisions.objectValues())
            for collection in collections:
                query = list(collection.query)
                for criterion in query:
                    # make sure 'review_state' value is a list
                    if criterion['i'] == 'review_state' and not isinstance(criterion['v'], list):
                        updateCollectionCriterion(collection, 'review_state', list(criterion['v']))
                    # make sure 'getDate' value is an unicode, not an integer
                    if criterion['i'] == 'getDate' and isinstance(criterion['v'], int):
                        updateCollectionCriterion(collection, 'getDate', unicode(criterion['v']))
        logger.info('Done.')

    def _removeUsersGlobalRoles(self):
        """Users should not have any global roles other than 'Member' and 'Authenticated',
           every other roles are given thru groups."""
        logger.info('Removing useless global roles for every users...')
        for member in api.user.get_users():
            roles = member.getRoles()
            useless_roles = [role for role in roles if role not in ['Member', 'Authenticated']]
            api.user.revoke_roles(user=member, roles=useless_roles)
        logger.info('Done.')

    def _updateItemColumnsKeys(self):
        """Some keys changed for static infos and groupsInCharge related fields in MeetingConfig.itemColumns,
           MeetingConfig.availableItemsListVisibleColumns and MeetingConfig.itemsListVisibleColumns."""
        logger.info('Updating MeetingConfig.itemColumns, MeetingConfig.availableItemsListVisibleColumns '
                    'and MeetingConfig.itemsListVisibleColumns...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            for field_name in ('itemColumns', 'availableItemsListVisibleColumns', 'itemsListVisibleColumns'):
                field = cfg.getField(field_name)
                keys = field.get(cfg)
                adapted_keys = []
                for k in keys:
                    if k in ['labels', 'item_reference', 'budget_infos']:
                        k = 'static_{0}'.format(k)
                    elif k == 'getGroupInCharge':
                        k = 'getGroupsInCharge'
                    elif k == 'group_in_charge_acronym':
                        k = 'groups_in_charge_acronym'
                    adapted_keys.append(k)
                field.set(cfg, adapted_keys)
        logger.info('Done.')

    def _adaptInternalImagesLinkToUseResolveUID(self):
        """We make sure we use resolveuid in src to internal images."""
        logger.info('Adapting link to internal images to use resolveuid...')
        # base our work on found images
        brains = self.portal.portal_catalog(portal_type='Image')
        i = 1
        total = len(brains)
        number_of_migrated_links = 0
        for brain in brains:
            logger.info('Migrating links to image {0}/{1} ({2})...'.format(
                i,
                total,
                brain.getPath()))
            i = i + 1
            image = brain.getObject()
            container = image.aq_inner.aq_parent
            # make sure image is added to meeting/item
            if container.meta_type not in ('Meeting', 'MeetingItem'):
                continue
            # get image url taking env var PUBLIC_URL into account
            image_url = get_public_url(image)
            image_UID = image.UID()
            for field in container.Schema().filterFields(default_content_type='text/html'):
                content = field.getRaw(container)
                if content.find(image_url) != -1:
                    content = content.replace(image_url, 'resolveuid/{0}'.format(image_UID))
                    logger.info('Replaced image link in field {0}'.format(field.getName()))
                    number_of_migrated_links = number_of_migrated_links + 1
                    field.set(container, content)
        logger.info('Adapted {0} links.'.format(number_of_migrated_links))
        logger.info('Done.')

    def _migrateContactPersonsKlass(self):
        """klass used by 'person' portal_type changed, this is only relevant for
           users using beta versions..."""
        logger.info('Migrating klass of collective.contact Person to PMPerson...')
        for brain in self.catalog(portal_type='person'):
            person = brain.getObject()
            if not isinstance(person, PMPerson):
                logger.info('Migrating person at {0}'.format('/'.join(person.getPhysicalPath())))
                migrate_base_class_to_new_class(
                    person,
                    old_class_name='collective.contact.core.content.person.Person',
                    new_class_name='Products.PloneMeeting.content.person.PMPerson')
        logger.info('Done.')

    def _disableVotes(self):
        """Disable the votes functionnality that is broken since we use contacts."""
        logger.info('Disabling votes for every MeetingConfigs...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.setUseVotes(False)
        logger.info('Done.')

    def _migrate_searchitemstoprevalidate_query(self):
        """Migrate query of the 'searchitemstoprevalidate' collection."""
        logger.info('Migrating query for collection searchitemstoprevalidate for every MeetingConfigs...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            searchitemstoprevalidate = cfg.searches.searches_items.get('searchitemstoprevalidate')
            if searchitemstoprevalidate:
                searches_infos = cfg._searchesInfo()
                searchitemstoprevalidate.query = list(searches_infos['searchitemstoprevalidate']['query'])
        logger.info('Done.')

    def _migrateItemsInConfig(self):
        """Migrate every items stored in MeetingConfig so provides IConfigElement."""
        logger.info('Migrating every items stored in MeetingConfig to provide IConfigElement...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            brains = api.content.find(context=cfg, meta_type='MeetingItem')
            for brain in brains:
                item = brain.getObject()
                if IConfigElement.providedBy(item):
                    self._already_migrated(False)
                    break
                alsoProvides(item, IConfigElement)
        logger.info('Done.')

    def _defaultFTWLabels(self):
        """To be overrided."""
        return {}

    def _initFTWLabels(self):
        """Initialize ftw labels and activate ones provided in env variable."""
        logger.info("Initializing ftw.labels...")
        defaultFTWLabels = self._defaultFTWLabels()
        if defaultFTWLabels:
            logger.info("Setting default ftw.labels...")
            defined_new_labels = False
            for cfg in self.tool.objectValues('MeetingConfig'):
                jar = ILabelJar(cfg)
                if not jar.storage:
                    defined_new_labels = True
                    update_labels_jar(jar, values=defaultFTWLabels)
            logger.info('Done.')

            # if some ftw_labels must be activated, we get it from env variable
            personal_labels = os.getenv('FTW_LABELS_PERSONAL_LABELS', [])
            if personal_labels and defined_new_labels:
                # personal_labels is a list as a string, so something like "['label']"
                personal_labels = literal_eval(personal_labels)
                logger.info("Initializing '${0}' personal labels...".format(
                    ', '.join(personal_labels)))
                for cfg in self.tool.objectValues('MeetingConfig'):
                    # if we activate some labels, then we enableLabels
                    cfg.setEnableLabels(True)
                    jar_storage = ILabelJar(cfg)
                    kept_personal_labels = [pl for pl in personal_labels
                                            if pl in jar_storage.storage]
                    if kept_personal_labels:
                        cfg._updatePersonalLabels(
                            personal_labels=kept_personal_labels, reindex=False)
            else:
                logger.info("No personal labels to activate...")

        logger.info('Done.')

    def _adaptShowHolidaysMessage(self):
        '''Show holidays_warning message to users having role Manager/MeetingManager,
           to avoid error message in log as now, tool/portal_plonemeeting is no more
           accessible to Anonymous.'''
        logger.info("Adapting required_roles for message 'Holidays warning'...")
        message = self.portal.get('messages-config').get('holidays_warning')
        if not message.required_roles:
            message.required_roles = ['Manager', 'MeetingManager']
        logger.info('Done.')

    def _migrateMeetingConfigDefaultAdviceHiddenDuringRedaction(self):
        """MeetingConfig.defaultAdviceHiddenDuringRedaction was a boolean, now it is a list."""
        logger.info('Migrating value for MeetingConfig.defaultAdviceHiddenDuringRedaction for every configs...')
        advice_portal_types = getAdvicePortalTypeIds()
        for cfg in self.tool.objectValues('MeetingConfig'):
            defaultAdviceHiddenDuringRedaction = cfg.defaultAdviceHiddenDuringRedaction
            if isinstance(defaultAdviceHiddenDuringRedaction, bool):
                if not defaultAdviceHiddenDuringRedaction:
                    cfg.setDefaultAdviceHiddenDuringRedaction([])
                else:
                    cfg.setDefaultAdviceHiddenDuringRedaction(advice_portal_types)
        logger.info('Done.')

    def _updateCatalogsByTypes(self):
        '''Before some types where blacklisted from portal_catalog, make sure it is no more.
           This was alredy done in migration to 4.0 but seems to be necessary again...'''
        logger.info('Make MeetingConfigs catalogued in portal_catalog...')
        at_tool = api.portal.get_tool('archetype_tool')
        at_tool.setCatalogsByType('MeetingConfig', ['portal_catalog'])
        # reindex every MeetingConfigs
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.reindexObject()
        logger.info('Done.')

    def run(self, extra_omitted=[]):
        logger.info('Migrating to PloneMeeting 4.1...')

        # recook CSS as we moved to Plone 4.3.18 and portal_css.concatenatedresources
        # could not exist, it is necessary for collective.js.tooltispter upgrade step
        try:
            self.portal.portal_css.concatenatedresources
        except AttributeError:
            self.portal.portal_css.cookResources()

        # upgrade imio.dashboard first as it takes care of migrating certain
        # profiles in particular order
        self._upgradeImioDashboard()

        self._migrateMeetingConfigDefaultAdviceHiddenDuringRedaction()
        self._updateCatalogsByTypes()
        self._migrateGroupsInChargeAttributes()

        # omit Products.PloneMeeting for now or it creates infinite loop as we are
        # in a Products.PloneMeeting upgrade step...
        self.upgradeAll(omit=['Products.PloneMeeting:default',
                              self.profile_name.replace('profile-', '')] + extra_omitted)

        # reinstall so versions are correctly shown in portal_quickinstaller
        # plone.app.versioningbehavior is installed
        self.reinstall(profiles=['profile-Products.PloneMeeting:default', ],
                       ignore_dependencies=False,
                       dependency_strategy=DEPENDENCY_STRATEGY_NEW)

        # Reload documentgenerator configuration from file
        update_oo_config()

        # power observers are now defined in MeetingConfig.powerObservers
        self._migratePowerObservers()
        # enable 'refused' WFAdadaption before reinstalling if relevant
        self._enableRefusedWFAdaptation()
        if self.profile_name != 'profile-Products.PloneMeeting:default':
            self.reinstall(profiles=[self.profile_name, ],
                           ignore_dependencies=False,
                           dependency_strategy=DEPENDENCY_STRATEGY_NEW)

        self.removeUnusedIndexes(indexes=['getTitle2', 'indexUsages', 'getGroupInCharge'])
        self.removeUnusedColumns(columns=['getTitle2', 'getRemoteUrl' 'getGroupInCharge'])
        # install collective.js.tablednd
        self.upgradeDependencies()
        self.cleanRegistries()
        # holidays 2020 were added
        self.updateHolidays()
        self.addNewSearches()

        # migration steps
        self._updateFacetedFilters()
        self._addItemTemplatesManagersGroup()
        # Update collections columns as column 'check_box_item' was renamed to 'select_row'
        self.updateCollectionColumns()
        self._markSearchesFoldersWithIBatchActionsMarker()
        self._removeMCPortalTabs()
        self._manageContentsKeptWhenItemSentToOtherMC()
        self._fixAnnexesMimeType()
        self._fixPODTemplatesMimeType()
        self._fixItemsWorkflowHistoryType()
        self._migrateToDoListSearches()
        self._adaptForContacts()
        self._adaptForPlonegroup()
        # update TAL conditions
        self.updateTALConditions(old_word='getPloneGroupsForUser', new_word='get_plone_groups_for_user')
        self.updateTALConditions(old_word='getGroupsForUser', new_word='get_orgs_for_user')
        self.updateTALConditions(old_word='omittedSuffixes', new_word='omitted_suffixes')
        self.updateTALConditions(old_word='isRestricted=True',
                                 new_word="power_observer_type='restrictedpowerobservers'")
        self.updateTALConditions(old_word='isRestricted=False',
                                 new_word="power_observer_type='powerobservers'")
        self.updateTALConditions(old_word='context.portal_plonemeeting',
                                 new_word='tool')
        self.updateTALConditions(old_word='here.portal_plonemeeting',
                                 new_word='tool')
        self.updateTALConditions(old_word='context.portal_plonemeeting.getMeetingConfig(context)',
                                 new_word='cfg')
        self.updateTALConditions(old_word='here.portal_plonemeeting.getMeetingConfig(here)',
                                 new_word='cfg')
        self.updateTALConditions(old_word='context.portal_plonemeeting.getMeetingConfig(here)',
                                 new_word='cfg')
        self.updateTALConditions(old_word='here.portal_plonemeeting.getMeetingConfig(context)',
                                 new_word='cfg')
        self.updateTALConditions(old_word='tool.getMeetingConfig(context)',
                                 new_word='cfg')
        self.updateTALConditions(old_word='tool.getMeetingConfig(here)',
                                 new_word='cfg')
        self._updateUsedAttributes()
        self._updateHistorizedAttributes()
        self._migrateGroupsShownInDashboardFilter()
        self._enableStyleTemplates()
        self.cleanMeetingConfigs(field_names=['defaultMeetingItemMotivation'])
        self.cleanTool(field_names=['extractTextFromFiles', 'availableOcrLanguages',
                                    'defaultOcrLanguage', 'enableUserPreferences'])
        self._fixMeetingCollectionsQuery()
        self._removeUsersGlobalRoles()
        self._updateItemColumnsKeys()
        self._adaptInternalImagesLinkToUseResolveUID()
        self._migrateContactPersonsKlass()
        self._disableVotes()
        self.removeUnusedPortalTypes(portal_types=['MeetingUser', 'MeetingFile',
                                                   'MeetingFileType', 'MeetingGroup'])
        self._migrate_searchitemstoprevalidate_query()
        self._migrateItemsInConfig()
        self._initFTWLabels()
        self._adaptShowHolidaysMessage()
        # too many indexes to update, rebuild the portal_catalog, but do not update other catalogs
        meeting_wf_ids = self.getWorkflows(meta_types=['Meeting'])
        self.refreshDatabase(workflows=True,
                             workflowsToUpdate=['plonemeeting_onestate_workflow'] + meeting_wf_ids,
                             catalogsToUpdate=[])


# The migration function -------------------------------------------------------
def migrate(context):
    '''This migration function will:

       1) Upgrade imio.dashboard;
       2) Make sure MeetingConfig is taken into account by portal_catalog;
       3) Rename MeetingConfig.itemGroupInChargeStates to MeetingConfig.itemGroupsInChargeStates;
       4) Upgrade all others packages;
       5) Reinstall PloneMeeting and upgrade dependencies;
       6) Make sure collective.documentgenerator OO parameters are correct;
       7) Migrate powerobservers;
       8) Enable 'refused' WF adaptation;
       9) Reinstall plugin if not PloneMeeting;
       10) Run common upgrades (dependencies, clean registries, holidays, reindexes);
       11) Add new faceted filters;
       12) Add '_itemtemplatesmanagers' groups;
       13) Update collections columns as column 'check_box_item' was renamed to 'select_row';
       14) Synch searches to mark searches sub folders with the IBatchActionsMarker;
       15) Remove MeetingConfig tabs from portal_actions portal_tabs;
       16) Migrate MeetingConfig.keepAdvicesOnSentToOtherMC to MeetingConfig.contentsKeptOnSentToOtherMC;
       17) Fix annex mimetype in case there was a problem with old annexes using uncomplete mimetypes_registry;
       18) Fix POD template mimetype because we need it to be correct for the styles template;
       19) Make sure workflow_history stored on items is a PersistentMapping;
       20) Migrate MeetingConfig.toDoListSearches as it is no more a ReferenceField;
       21) Adapt application for Contacts;
       22) Update MeetingConfig.usedItemAttributes, select 'description' and unselect 'itemAssembly';
       23) Migrate MeetingConfig.groupsShownInDashboardFilter to MeetingConfig.groupsHiddenInDashboardFilter;
       24) Configure MeetingConfig podtemplates folder to accept style templates;
       25) Clean MeetingConfigs from removed attributes;
       26) Fix meeting related DashboardCollections query;
       27) Remove global roles for every users, roles are only given thru groups;
       28) Update keys stored in MeetingConfig related to static infos displayed in dashboards;
       29) Adapt link to images to be sure to use resolveuid;
       30) Migrate contact person klass to use PMPerson for users of beta versions...;
       31) Disable votes functionnality;
       32) Removed no more used portal_types;
       33) Migrate 'searchitemstoprevalidate' query;
       34) Migrate items in MeetingConfig so it provides IConfigElement;
       35) Initialize personal labels if found in FTW_LABELS_PERSONAL_LABELS env variable;
       36) Adapt 'Show holidays warning' condition to not evaluate it if user is
           Anonymous because access to tool is no more granted to Anonymous;
       37) Migrate MeetingConfig.defaultAdviceHiddenDuringRedaction (bool to list);
       38) Invalidate all cache before cataloging and returning to application
       39) Refresh catalog and types using workflows of plonemeeting_onestate_workflow
          (no more access to Anonymous) and Meeting workflows (MeetingManager have 'Review portal content'
          in state 'closed').
    '''
    migrator = Migrate_To_4_1(context)
    migrator.run()
    migrator.finish()
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from Products.Five import BrowserView
from Products.PloneMeeting.interfaces import IRedirect
from zope.component.hooks import getSite
from zope.interface import implements


class RedirectView(BrowserView):
    """
    """
    DEFAULT = """<html><head><base href="{0}" /></head></html>"""

    def __call__(self):
        """ """
        url = self.request.form.get('url', '')
        if 'ajax_load' in self.request.form:
            return self.DEFAULT.format(url)
        else:
            self.request.RESPONSE.redirect(url)


class Redirect(object):
    """
      Redirect to the right place, this is necessary for overlays to work correctly with z3c.form...
    """
    implements(IRedirect)

    def __init__(self, request):
        self.request = request
        self.ajax_load = self.request.form.get('ajax_load', '')

    def redirect(self, url):
        """
          See docstring in interfaces.IRedirect
        """
        portal = getSite()
        result = []
        result.append(portal.absolute_url())
        result.append("/@@redirect_view?")
        if self.ajax_load:
            result.append("ajax_load=%s&" % self.ajax_load)
        result.append("url=%s" % url)
        self.request.RESPONSE.redirect(''.join(result))



# -*- coding: utf-8 -*-
#
# File: testPerformances.py
#
# GNU General Public License (GPL)
#

from collective.contact.plonegroup.utils import get_organization
from collective.contact.plonegroup.utils import get_organizations
from collective.contact.plonegroup.utils import get_plone_groups
from collective.eeafaceted.batchactions.utils import listify_uids
from datetime import datetime
from datetime import timedelta
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import invalidate_cachekey_volatile_for
from imio.helpers.content import get_user_fullname
from plone import api
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.utils import getToolByName
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from Products.PloneMeeting.tests.PloneMeetingTestCase import pm_logger
from Products.PloneMeeting.utils import down_or_up_wf
from Products.PloneMeeting.utils import get_annexes
from Products.PloneMeeting.utils import isPowerObserverForCfg
from profilehooks import timecall
from zope.event import notify


class testPerformances(PloneMeetingTestCase):
    '''Tests various aspects of performances.'''

    def setUp(self):
        # call parent setUp
        PloneMeetingTestCase.setUp(self)

    def _setupForDelayingItems(self, number_of_items, number_of_annexes):
        """ """
        meeting, uids = self._setupMeetingItemsWithAnnexes(number_of_items,
                                                           number_of_annexes,
                                                           present_items=True)
        # set the meeting in the 'decided' state
        self.decideMeeting(meeting)
        # in some wfs, deciding a meeting will accept every items...
        # set back items to the 'itemfrozen' state
        for itemInMeeting in meeting.get_items():
            if itemInMeeting.query_state() == 'itemfrozen':
                break
            self.do(itemInMeeting, 'backToItemFrozen')

        return meeting, uids

    def _setupMeetingItemsWithAnnexes(self,
                                      number_of_items,
                                      number_of_annexes,
                                      with_meeting=True,
                                      present_items=False,
                                      as_uids=True):
        cfg = self.meetingConfig
        wfAdaptations = list(cfg.getWorkflowAdaptations())
        if 'no_publication' not in wfAdaptations:
            self.changeUser('siteadmin')
            wfAdaptations.append('no_publication')
            cfg.setWorkflowAdaptations(wfAdaptations)
            notify(ObjectEditedEvent(cfg))

        self.changeUser('pmManager')
        meeting = None
        if with_meeting:
            # create a meeting
            meeting = self.create('Meeting')
        data = {}
        uids = []
        items = []
        logger_threshold = 10
        created_items = 0
        total_created_items = 0
        for i in range(number_of_items):
            # display message in logger while creating many items
            created_items += 1
            total_created_items += 1
            if created_items == logger_threshold:
                if not number_of_annexes:
                    pm_logger.info('Created %d out of %d items' % (total_created_items, number_of_items))
                created_items = 0

            # create the item
            data['title'] = 'Item number %d' % i
            item = self.create('MeetingItem', **data)
            item.setDecision('<p>A decision</p>')
            # add annexes
            if number_of_annexes:
                for j in range(number_of_annexes):
                    self.addAnnex(item, annexTitle="Annex number %d" % j)
                pm_logger.info('Added %d annexes to the item number %s' %
                               (number_of_annexes, total_created_items))
            if present_items:
                self.presentItem(item)
            uids.append(item.UID())
            items.append(item)
        if as_uids:
            return meeting, uids
        else:
            return meeting, items

    def test_pm_Delay5ItemsWith0Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(5, 0)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (5, 0))
        self._delaySeveralItems(meeting, uids)

    def test_pm_Delay10ItemsWith0Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(10, 0)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (10, 0))
        self._delaySeveralItems(meeting, uids)

    def test_pm_Delay5ItemsWith5Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(5, 5)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (5, 5))
        self._delaySeveralItems(meeting, uids)

    def test_pm_Delay10ItemsWith5Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(10, 5)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (10, 5))
        self._delaySeveralItems(meeting, uids)

    def test_pm_Delay5ItemsWith10Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(5, 10)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (5, 10))
        self._delaySeveralItems(meeting, uids)

    def test_pm_Delay10ItemsWith10Annexes(self):
        '''While delaying an item, it is cloned with annexes.'''
        meeting, uids = self._setupForDelayingItems(10, 10)
        pm_logger.info('Delay %d items containing %d annexes in each.' % (10, 10))
        self._delaySeveralItems(meeting, uids)

    def _setItemReferenceFormat(self):
        """Compute metingDate, proposingGroup acronym and item number relativeTo meeting."""
        self.meetingConfig.setItemReferenceFormat(
            "python: here.restrictedTraverse('pm_unrestricted_methods').getLinkedMeetingDate().strftime('%Y%m%d') + "
            "'/' + here.getProposingGroup(True).getAcronym() + '/' + "
            "str(here.getItemNumber(relativeTo='meeting', for_display=True))")

    def test_pm_Update250ItemsItemReference(self):
        '''Update the itemReference of 250 items.'''
        self._setItemReferenceFormat()
        meeting, uids = self._setupMeetingItemsWithAnnexes(250, 0, with_meeting=True, present_items=True)

        # item references are only updated once meeting is frozen
        # freeze meeting but defer references update
        # update every items
        pm_logger.info('Freezing meeting without update_item_references.')
        self.request.set('defer_Meeting_update_item_references', True)
        self.freezeMeeting(meeting)
        self.request.set('defer_Meeting_update_item_references', False)
        # update every items
        pm_logger.info(
            'Updating item references for %d items presented in a meeting starting from item number %s.' % (250, 0))
        self._updateItemReferences(meeting)
        # update items starting from 100th item
        pm_logger.info(
            'Updating item references for %d items presented in a meeting starting from item number %s.' % (250, 100))
        self._updateItemReferences(meeting, start_number=100 * 100)
        # update items starting from 200th item
        pm_logger.info(
            'Updating item references for %d items presented in a meeting starting from item number %s.' % (250, 200))
        self._updateItemReferences(meeting, start_number=200 * 100)

    @timecall
    def _updateItemReferences(self, meeting, start_number=0):
        '''Helper method that actually compute every items itemReference for p_meeting.'''
        # set back every items reference to '' so the entire process including reindex of SearchableText is done
        for item in meeting.get_items():
            item.setItemReference('')
        meeting.update_item_references(start_number=start_number)

    def test_pm_Present50ItemsWithoutAnnexesSeveralTimes(self):
        '''While presenting items, these items are inserted in a given order.
           In this test, as every items use same 'proposingGroup', same 'privacy'
           and same 'listType' every items are evaluated each time and
           every new is finally added at the end of the meeting.
           We present 50 by 50 items successively in same meeting'''
        pm_logger.info('Presenting %d items without annexes in a meeting containing %d items.' % (50, 0))
        # use 'complex' inserting method
        self.meetingConfig.setInsertingMethodsOnAddItem(({'insertingMethod': 'on_list_type',
                                                          'reverse': '0'},
                                                         {'insertingMethod': 'on_privacy',
                                                          'reverse': '0'},
                                                         {'insertingMethod': 'on_proposing_groups',
                                                          'reverse': '0'},))
        meeting, items = self._setupMeetingItemsWithAnnexes(50, 0, as_uids=False)
        # called when no item in the meeting
        self._presentSeveralItems(items)
        # called second times whith items in the meeting
        pm_logger.info('Presenting %d items without annexes in a meeting containing %d items.'
                       % (50, meeting.number_of_items()))
        dummy_meeting, items = self._setupMeetingItemsWithAnnexes(50, 0, with_meeting=False, as_uids=False)
        self._presentSeveralItems(items)
        # called third times whith items in the meeting
        pm_logger.info('Presenting %d items without annexes in a meeting containing %d items.'
                       % (50, meeting.number_of_items()))
        dummy_meeting, items = self._setupMeetingItemsWithAnnexes(50, 0, with_meeting=False, as_uids=False)
        self._presentSeveralItems(items)
        # called fourth times whith items in the meeting
        pm_logger.info('Presenting %d items without annexes in a meeting containing %d items.'
                       % (50, meeting.number_of_items()))
        dummy_meeting, items = self._setupMeetingItemsWithAnnexes(50, 0, with_meeting=False, as_uids=False)
        self._presentSeveralItems(items)

    def test_pm_SendSeveralItemsWithAnnexesToAnotherMC(self):
        '''We will freeze a meeting containing 50 items from which 25 need to be send
           to another MC.  Every items contain 5 annexes.'''
        pm_logger.info('Freezing a meeting containing %d items and sending %d items to another MC.' % (50, 25))
        cfg = self.meetingConfig
        self._enableField('category', enable=False)
        cfg.setInsertingMethodsOnAddItem(({'insertingMethod': 'on_proposing_groups',
                                           'reverse': '0'}, ))
        cfg2 = self.meetingConfig2
        self._enableField('category', cfg=cfg2, enable=False)
        cfg2.setInsertingMethodsOnAddItem(({'insertingMethod': 'on_proposing_groups',
                                            'reverse': '0'}, ))
        cfg2Id = cfg2.getId()
        # make items sent to config2 automatically presented in the next meeting
        cfg.setMeetingConfigsToCloneTo(({'meeting_config': '%s' % cfg2Id,
                                         'trigger_workflow_transitions_until': '%s.%s' %
                                         (cfg2Id, 'present')},))
        cfg.setItemAutoSentToOtherMCStates((u'itemfrozen', ))
        meeting, items = self._setupMeetingItemsWithAnnexes(50, 5, present_items=True, as_uids=False)
        # make 25 items sendable to another MC
        for item in items[0:25]:
            item.setOtherMeetingConfigsClonableTo((self.meetingConfig2.getId(), ))
            item.reindexObject(idxs=['sentToInfos', ])

        # create meeting in cfg2 in which items will be presented
        self.setMeetingConfig(cfg2Id)
        meeting2 = self.create('Meeting', date=datetime.now() + timedelta(days=1))
        self.assertFalse(meeting2.get_items())

        # freeze the meeting, this will do the job
        self._freezeMeetingAndSendItemsToAnotherMC(meeting)

        # make sure meeting2 has items
        self.assertEquals(len(meeting2.get_items()), 25)

    @timecall
    def _delaySeveralItems(self, meeting, uids):
        '''Helper method that actually delays the items.'''
        brains = api.content.find(UID=uids)
        for brain in brains:
            obj = brain.getObject()
            api.content.transition(obj=obj, transition='delay')

    @timecall
    def _presentSeveralItems(self, items):
        '''Present the p_items in p_meeting.'''
        for item in items:
            self.presentItem(item)

    @timecall
    def _freezeMeetingAndSendItemsToAnotherMC(self, meeting):
        '''Freeze given p_meeting.'''
        self.freezeMeeting(meeting)

    def test_pm_ComputeItemNumberWithSeveralNotClosedMeetings(self):
        '''Check performances while looking for the current item number using
           MeetingItem.getItemNumber(relativeTo='meetingConfig') that will query previous
           existing meetings to get the item number.'''
        self.changeUser('pmManager')
        # create 30 meetings containing 150 items in each
        data = {}
        meetings = []
        number_of_meetings = 5
        number_of_items = 10
        pm_logger.info('Adding %d meetings with %d items in each' % (number_of_meetings, number_of_items))
        for i in range(number_of_meetings):
            pm_logger.info('Creating meeting %d of %s' % (i + 1, number_of_meetings))
            meeting = self.create('Meeting', date=datetime.now() + timedelta(i + 1))
            meetings.append(meeting)
            for j in range(number_of_items):
                data['title'] = 'Item number %d' % j
                item = self.create('MeetingItem', **data)
                item.setDecision('<p>A decision</p>')
                # present the item
                self.presentItem(item)
        # now we have number_of_meetings meetings containing number_of_items items
        # test with the last created meeting
        self._computeItemNumbersForMeeting(meeting)
        # now close meeting at half of existing meetings
        meetingAtHalf = meetings[int(number_of_meetings / 2)]
        self.closeMeeting(meetingAtHalf)
        self._computeItemNumbersForMeeting(meeting)
        # now close meeting at 90% of created meetings that is the most obvious usecase
        meetingAt90Percent = meetings[int(number_of_meetings * 0.9)]
        self.closeMeeting(meetingAt90Percent)
        self._computeItemNumbersForMeeting(meeting)
        # now close penultimate meeting (the meeting just before the meeting the item is in) and test again
        meetingPenultimate = meetings[number_of_meetings - 2]
        self.closeMeeting(meetingPenultimate)
        self._computeItemNumbersForMeeting(meeting)
        # finally close the meeting we will compute items numbers
        self.closeMeeting(meeting)
        self._computeItemNumbersForMeeting(meeting)

    @timecall
    def _computeItemNumbersForMeeting(self, meeting):
        '''Helper method that actually compute item number for every items of the given p_meeting.'''
        for item in meeting.get_items():
            item.getItemNumber(relativeTo='meetingConfig')

    def _setupForOrgs(self, number_of_orgs):
        self.changeUser('admin')
        cfg = self.meetingConfig
        cfg2 = self.meetingConfig2
        # remove existing groups and add our own
        # make what necessary for groups to be removable...
        cfg.setOrderedGroupsInCharge(())
        cfg.setSelectableCopyGroups(())
        cfg.setSelectableAdvisers(())
        cfg2.setOrderedGroupsInCharge(())
        cfg2.setSelectableCopyGroups(())
        cfg2.setSelectableAdvisers(())
        orgs = get_organizations(only_selected=True)
        for org in orgs:
            self._select_organization(org.UID(), remove=True)
            for ploneGroup in get_plone_groups(org.UID()):
                for memberId in ploneGroup.getGroupMemberIds():
                    ploneGroup.removeMember(memberId)
        # remove items defined in the tool
        self._removeConfigObjectsFor(cfg, folders=['recurringitems', 'itemtemplates', 'categories'])
        self._removeConfigObjectsFor(cfg2, folders=['recurringitems', 'itemtemplates', 'categories'])

        # remove groups
        self._removeOrganizations()
        # create groups
        for i in range(number_of_orgs):
            org = self.create('organization', id=i, title='Org %d' % i)
            self._select_organization(org.UID())

    def test_pm_get_organizations_caching(self):
        '''Test collective.contact.plonegroup.utils.get_organizations caching.'''
        for n in [10, 100, 250]:
            # first test with 10 orgs
            self._setupForOrgs(n)
            pm_logger.info('get_organizations called 1 time with %d activated orgs.' % n)
            # first call, even when asking caching, it is not as never computed but it is cached
            pm_logger.info('No caching.')
            self._get_organizations(times=1, caching=True)
            # second time, cached
            pm_logger.info('Caching.')
            self._get_organizations(times=1, caching=True)
            # remove cache
            self.cleanMemoize()
            pm_logger.info('get_organizations called 100 times with %d activated orgs.' % n)
            pm_logger.info('No caching.')
            self._get_organizations(times=100, caching=False)
            # second time, cached
            pm_logger.info('Caching.')
            self._get_organizations(times=100, caching=True)
            # remove cache
            self.cleanMemoize()

    @timecall
    def _get_organizations(self, times=1, caching=True):
        ''' '''
        for time in range(times):
            get_organizations(not_empty_suffix='advisers', caching=caching)

    def test_pm_get_organization_caching(self):
        '''Test collective.contact.plonegroup.utils.get_organization caching.
           Check if faster than using catalog unrestricted.'''
        # test with 100 orgs
        self._setupForOrgs(100)
        pm_logger.info('get_organization called 1000 time with 100 activated orgs.')
        # with caching=False
        pm_logger.info('No caching.')
        self._get_organization(times=1000, caching=False)
        # with caching=True
        pm_logger.info('Caching.')
        self._get_organization(times=1000, caching=True)

    @timecall
    def _get_organization(self, times=1, caching=True):
        ''' '''
        for time in range(times):
            get_organization(self.vendors_uid, caching=caching)

    def test_pm_SpeedGetMeetingConfig(self):
        '''Test ToolPloneMeeting.getMeetingConfig method performances.
           We call the method 2000 times, this is what happens when displaying
           a meeting containing 100 items.'''
        # create an item
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        # call getMeetingConfig 2000 times with item
        self._getMeetingConfigOnTool(item, 2000)
        # call getMeetingConfig 2000 times with item parent
        self._getMeetingConfigOnTool(item.aq_inner.aq_parent, 2000)
        # call getMeetingConfig 2000 times with searches_items folder
        self._getMeetingConfigOnTool(item.aq_inner.aq_parent.searches_items, 2000)

    @timecall
    def _getMeetingConfigOnTool(self, context, times=1):
        ''' '''
        for time in range(times):
            self.tool.getMeetingConfig(context)

    def test_pm_SpeedSetManuallyLinkedItems(self):
        '''Test MeetingItem.setManuallyLinkedItems method performances.'''

        meeting1, items1 = self._setupMeetingItemsWithAnnexes(10, 0, as_uids=False)
        meeting2, items2 = self._setupMeetingItemsWithAnnexes(10, 0, as_uids=False)
        meeting3, items3 = self._setupMeetingItemsWithAnnexes(10, 0, as_uids=False)
        meeting4, items4 = self._setupMeetingItemsWithAnnexes(10, 0, as_uids=False)
        meeting5, items5 = self._setupMeetingItemsWithAnnexes(10, 0, as_uids=False)
        item = items1[0]
        linked_items = items1[1:] + items2 + items3 + items4 + items5
        linked_item_uids = [linked_item.UID() for linked_item in linked_items]
        # set with caching
        pm_logger.info('setManuallyLinkedItems with cahing=True (1).')
        self._setManuallyLinkedItemsOnItem(item, linked_item_uids)
        # set without caching
        item.setManuallyLinkedItems([])
        pm_logger.info('setManuallyLinkedItems with cahing=False (1).')
        self._setManuallyLinkedItemsOnItem(item, linked_item_uids, caching=False)
        # set with caching second time
        item.setManuallyLinkedItems([])
        pm_logger.info('setManuallyLinkedItems with cahing=True (2).')
        self._setManuallyLinkedItemsOnItem(item, linked_item_uids)
        # set without caching second time
        item.setManuallyLinkedItems([])
        pm_logger.info('setManuallyLinkedItems with cahing=False (2).')
        self._setManuallyLinkedItemsOnItem(item, linked_item_uids, caching=False)

    @timecall
    def _setManuallyLinkedItemsOnItem(self, item, linked_item_uids, caching=True):
        ''' '''
        item.setManuallyLinkedItems(linked_item_uids, caching=caching)

    def test_pm_SpeedIsManager(self):
        '''Test ToolPloneMeeting.isManager method performances.
           Need to comment manually ram.cache on ToolPloneMeeting.isManager
           to see difference.  Performance is better with @ram.cache.'''
        # create an item
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        self._isManager(item, 2000)

    @timecall
    def _isManager(self, context, times=1):
        ''' '''
        for time in range(times):
            self.tool.isManager(self.meetingConfig)
            self.tool.isManager(realManagers=True)

    def test_pm_GetAuthenticatedMember(self):
        '''Test performance between portal_membership.getAuthenticatedMember and
           plone_portal_state.member() and api.user.get_current.'''
        # create an item
        self.changeUser('pmManager')
        # call getAuthenticatedMember 2000 times
        self._portalMembershipGetAuthenticatedMember(2000)
        # call plone_portal_state.member() 2000 times
        self._plonePortalStateMember(2000)
        # call api.user.get_current() 2000 times
        self._apiUserGetCurrent(2000)
        # call utils.get_current_user_id() 2000 times with
        # AUTHENTICATED_USER set
        self._utilsGetCurrentUserId(2000)
        self.request.set("AUTHENTICATED_USER", None)
        # call utils.get_current_user_id() 2000 times with
        # AUTHENTICATED_USER not set, will use api.user.get_current()
        self._utilsGetCurrentUserId(2000)

    @timecall
    def _portalMembershipGetAuthenticatedMember(self, times=1):
        ''' '''
        for time in range(times):
            getToolByName(self.portal, 'portal_membership').getAuthenticatedMember()

    @timecall
    def _plonePortalStateMember(self, times=1):
        ''' '''
        for time in range(times):
            self.portal.restrictedTraverse('@@plone_portal_state').member()

    @timecall
    def _apiUserGetCurrent(self, times=1):
        ''' '''
        for time in range(times):
            api.user.get_current()

    @timecall
    def _utilsGetCurrentUserId(self, times=1):
        ''' '''
        for time in range(times):
            get_current_user_id()

    def _setupForMeetingCategories(self, number_of_categories, withUsingGroups=False):
        self.changeUser('admin')
        # remove items in the tool and categories
        self._removeConfigObjectsFor(self.meetingConfig, folders=['recurringitems', 'itemtemplates', 'categories'])
        # create categories
        for i in range(number_of_categories):
            catObj = self.create('meetingcategory',
                                 id=i,
                                 title='Category %d' % i)
            if withUsingGroups:
                catObj.setUsingGroups(('developers', ))
            catObj._at_creation_flag = False
            catObj.at_post_create_script()

    def test_pm_GetCategoriesCaching(self):
        '''Test MeetingConfig.getCategories caching.'''
        times = 1000
        self._enableField('category')
        # first test with 10 groups without usingGroups
        self._setupForMeetingCategories(10, withUsingGroups=False)
        pm_logger.info('getCategories called %d times with %d activated groups, without usingGroups.'
                       % (times, 10))
        pm_logger.info('Not cached..')
        self._getCategoriesOnMeetingConfig(times=times)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

        # first test with 10 groups with usingGroups
        self._setupForMeetingCategories(10, withUsingGroups=True)
        pm_logger.info('getCategories called %d times with %d activated groups, with usingGroups.'
                       % (times, 10))
        pm_logger.info('No cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

        # test with 100 categories without usingGroups
        self._setupForMeetingCategories(100, withUsingGroups=False)
        pm_logger.info('getCategories called %d times with %d activated groups, without usingGroups.'
                       % (times, 100))
        pm_logger.info('No cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

        # test with 100 categories with usingGroups
        self._setupForMeetingCategories(100, withUsingGroups=True)
        pm_logger.info('getCategories called %d times with %d activated groups, with usingGroups.'
                       % (times, 100))
        pm_logger.info('No cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

        # test with 250 categories without usingGroups
        self._setupForMeetingCategories(250, withUsingGroups=False)
        pm_logger.info('getCategories called %d times with %d activated groups, without usingGroups.'
                       % (times, 250))
        pm_logger.info('No cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=times)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

        # test with 250 categories with usingGroups
        self._setupForMeetingCategories(250, withUsingGroups=True)
        pm_logger.info('getCategories called %d times with %d activated groups, with usingGroups.'
                       % (times, 250))
        pm_logger.info('No cached.')
        self._getCategoriesOnMeetingConfig(times=1000)
        # second time, cached
        pm_logger.info('Cached.')
        self._getCategoriesOnMeetingConfig(times=1000)
        # remove cache
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")

    @timecall
    def _getCategoriesOnMeetingConfig(self, times=1):
        ''' '''
        for time in range(times):
            self.meetingConfig.getCategories(userId='pmManager', onlySelectable=True)

    def _setupItemsForUpdateLocalRoles(self, add_advices=True, add_annexes=True):
        '''Call.update_local_roles on items holding many annexes and advices.'''
        # configure several auto asked advices and manually asked advices
        cfg = self.meetingConfig
        self.changeUser('siteadmin')
        if add_advices:
            # create some groups in charge
            gic_uids = []
            for i in range(1, 11):
                gic = self.create(
                    'organization',
                    id='groupincharge{0}'.format(i),
                    Title='Group in charge {0}'.format(i),
                    acronym='GIC{0}'.format(i))
                gic_uids.append(gic.UID())
            # set the groups in charge, in charge of developers
            self.developers.groups_in_charge = gic_uids
            # configure customAdvisers
            custom_advisers = []
            for gic_uid in gic_uids:
                custom_advisers.append(
                    {'gives_auto_advice_on':
                        "python:'{0}' in item.getGroupsInCharge(fromOrgIfEmpty=True)".format(gic_uid),
                     'org': gic_uid,
                     'for_item_created_from': '2019/10/15',
                     'row_id': 'row_id__{0}'.format(gic_uid)})
            cfg.setCustomAdvisers(custom_advisers)

        self.changeUser('pmManager')
        # create 50 items with 20 annexes
        number_of_items = 50
        number_of_annexes = 0
        if add_annexes:
            number_of_annexes = 20
        meeting, items = self._setupMeetingItemsWithAnnexes(
            number_of_items=number_of_items,
            number_of_annexes=number_of_annexes,
            with_meeting=False,
            as_uids=False)
        return items

    def test_pm_UpdateLocalRolesOn50ItemsWith0AnnexesAnd0Advices(self):
        '''Call.update_local_roles on items without any annexes or advices.'''
        items = self._setupItemsForUpdateLocalRoles(add_advices=False,
                                                    add_annexes=False)
        number_of_advices = 0
        for item in items:
            self.assertEqual(len(item.adviceIndex), number_of_advices)

        # call update local roles 2 times
        uids = listify_uids([item.UID() for item in items])
        number_of_annexes = 0
        for item in items:
            self.assertEqual(len(get_annexes(item)), number_of_annexes)

        pm_logger.info(
            'Call 1 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)
        pm_logger.info(
            'Call 2 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)

    def test_pm_UpdateLocalRolesOn50ItemsWith20AnnexesAnd0Advices(self):
        '''Call.update_local_roles on items with 20 annexes and 0 advices.'''
        items = self._setupItemsForUpdateLocalRoles(add_advices=False,
                                                    add_annexes=True)
        number_of_advices = 0
        for item in items:
            self.assertEqual(len(item.adviceIndex), number_of_advices)

        # call update local roles 2 times
        uids = listify_uids([item.UID() for item in items])
        number_of_annexes = 20
        for item in items:
            self.assertEqual(len(get_annexes(item)), number_of_annexes)

        pm_logger.info(
            'Call 1 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)
        pm_logger.info(
            'Call 2 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)

    def test_pm_UpdateLocalRolesOn50ItemsWith0AnnexesAnd10Advices(self):
        '''Call.update_local_roles on items with 0 annexes and 10 advices.'''
        items = self._setupItemsForUpdateLocalRoles(add_advices=True,
                                                    add_annexes=False)
        number_of_advices = 10
        for item in items:
            self.assertEqual(len(item.adviceIndex), number_of_advices)

        # call update local roles 2 times
        uids = listify_uids([item.UID() for item in items])
        number_of_annexes = 0
        for item in items:
            self.assertEqual(len(get_annexes(item)), number_of_annexes)
        pm_logger.info(
            'Call 1 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)
        pm_logger.info(
            'Call 2 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)

    def test_pm_UpdateLocalRolesOn50ItemsWith20AnnexesAnd10Advices(self):
        '''Call.update_local_roles on items with 20 annexes and 10 advices.'''
        items = self._setupItemsForUpdateLocalRoles(add_advices=True,
                                                    add_annexes=True)
        number_of_advices = 10
        for item in items:
            self.assertEqual(len(item.adviceIndex), number_of_advices)

        # call update local roles 2 times
        uids = listify_uids([item.UID() for item in items])
        number_of_annexes = 20
        for item in items:
            self.assertEqual(len(get_annexes(item)), number_of_annexes)

        pm_logger.info(
            'Call 1 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)
        pm_logger.info(
            'Call 2 to.update_local_roles on 50 items holding '
            '{0} annexes and {1} auto asked advices.'.format(number_of_annexes, number_of_advices))
        self._updateItemLocalRoles(uids)

    @timecall
    def _updateItemLocalRoles(self, uids):
        '''Helper method that actually update local roles on items.'''
        self.tool.update_all_local_roles(**{'UID': uids})

    def test_pm_DuplicateItemWith50Annexes(self):
        '''Duplicate an item containing 50 annexes.'''
        dummy_meeting, items = self._setupMeetingItemsWithAnnexes(
            2, 50, with_meeting=False, as_uids=False)
        for item in items:
            # duplicate it 2 times
            pm_logger.info('Duplicate item first time without annexes.')
            self._duplicateItem(item)
            pm_logger.info('Duplicate item first time with annexes.')
            self._duplicateItem(item, copyAnnexes=True)
            pm_logger.info('Duplicate item second time without annexes.')
            self._duplicateItem(item)
            pm_logger.info('Duplicate item second time with annexes.')
            self._duplicateItem(item, copyAnnexes=True)

    @timecall
    def _duplicateItem(self, item, copyAnnexes=False):
        '''Helper method that actually duplicated given p_item.'''
        item.clone(copyAnnexes=copyAnnexes)

    def test_pm_SpeedGetMeeting(self):
        '''Test MeetingItem.getMeeting method performances.
           We call the method 2000 times, this is what happens when displaying
           a dashboard of 100 items.'''
        self.changeUser('pmManager')
        self.create('Meeting')
        item = self.create('MeetingItem')
        self.presentItem(item)
        # call getMeeting 2000 times wihout caching
        self._getMeetingOnItem(item, 2000, caching=False)
        # call getMeeting 2000 times with caching
        self._getMeetingOnItem(item, 2000, caching=True)

    @timecall
    def _getMeetingOnItem(self, item, times=1, caching=True):
        ''' '''
        pm_logger.info(
            'Call {0} times, with caching={1}'.format(times, caching))
        for time in range(times):
            item.getMeeting(only_uid=False, caching=caching)

    def test_pm_ToolGroupIsNotEmpty(self):
        '''Test ToolPloneMeeting.group_is_not_empty method performances.
           More performant without ram.cache'''
        self.changeUser('pmManager')
        # call group_is_not_empty 2000 times
        self._tool_group_is_not_empty(self.vendors_uid, "creators", times=1000)

    @timecall
    def _tool_group_is_not_empty(self, org_uid, suffix, times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            self.tool.group_is_not_empty(self.vendors_uid, "creators")

    def test_pm_SpeedGetMemberInfo(self):
        '''Test MembershipTool.getMemberInfo that is monkeypatched.
           This needs to be done manually, enable or disable ram.cache decorator.'''
        self.changeUser('pmManager')
        # call getMemberInfo 1000 times
        membership = api.portal.get_tool('portal_membership')
        self._get_member_info(membership, times=1000)

    @timecall
    def _get_member_info(self, membership, times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            membership.getMemberInfo("pmManager")

    def test_pm_SpeedGetUserFullname(self):
        '''Test imio.helpers.content.get_user_fullname.'''
        self.changeUser('pmManager')
        # call get_user_fullname 1000 times
        self._get_user_fullname(times=1000)

    @timecall
    def _get_user_fullname(self, userId="pmManager", times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            get_user_fullname(userId)

    def test_pm_SpeedItemQueryState(self):
        '''Test MeetingItem.query_state.'''
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        # get a large workflow_history
        self.validateItem(item)
        self.backToState(item, 'itemcreated')
        self.validateItem(item)
        self.backToState(item, 'itemcreated')
        self.validateItem(item)
        # call query_state 1000 times
        self._query_state(item, times=1000)

    @timecall
    def _query_state(self, item, times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            item.query_state()

    def test_pm_SpeedIsPowerObserverForCfg(self):
        '''Test utils.isPowerObserverForCfg.'''
        self.changeUser('pmManager')
        # call it 1000 times
        self._isPowerObserverForCfg(times=1000)

    @timecall
    def _isPowerObserverForCfg(self, times=1):
        ''' '''
        cfg = self.meetingConfig
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            isPowerObserverForCfg(cfg)
            isPowerObserverForCfg(cfg, ["powerobservers"])
            isPowerObserverForCfg(cfg, ["restrictedpowerobservers"])
            isPowerObserverForCfg(cfg, ["powerobservers", "restrictedpowerobservers"])

    def test_pm_SpeedUserIsAmong(self):
        '''Test ToolPloneMeeting.userIsAmong.'''
        self.changeUser('pmManager')
        # call it 1000 times
        self._userIsAmong(times=1000)

    @timecall
    def _userIsAmong(self, times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            self.tool.userIsAmong(["advisers"])
            self.tool.userIsAmong(["creators", "reviewers"])
            self.tool.userIsAmong(["powerobservers"])

    def test_pm_SpeedUtilsdown_or_up_wf(self):
        '''Test utils.down_or_up_wf for MeetingItem.'''
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        # get a larger workflow_history
        self.validateItem(item)
        self.backToState(item, 'itemcreated')
        self.validateItem(item)
        self.backToState(item, 'itemcreated')
        self.validateItem(item)
        # call down_or_up_wf 1000 times
        self._down_or_up_wf(item, times=1000)

    @timecall
    def _down_or_up_wf(self, item, times=1):
        ''' '''
        pm_logger.info('Call {0} times'.format(times))
        for time in range(times):
            down_or_up_wf(item)

    def test_pm_CheckAndHasPermission(self):
        '''Test _checkPermission and user.has_permission.'''
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        # call _checkPermission 1000 times
        self._check_permission(item, times=1000)
        # call user.has_permission 1000 times
        self._user_has_permission(item, times=1000)

    @timecall
    def _check_permission(self, item, times=1):
        ''' '''
        pm_logger.info('Call _checkPermission {0} times'.format(times))
        for time in range(times):
            _checkPermission("ModifyPortalContent", item)

    @timecall
    def _user_has_permission(self, item, times=1):
        ''' '''
        pm_logger.info('Call user.has_permission {0} times'.format(times))
        for time in range(times):
            self.member.has_permission("ModifyPortalContent", item)

    def test_pm_Speed_getLastValidationState(self):
        '''Test MeetingItemWorkflowConditions._getLastValidationState.'''
        self.changeUser('pmManager')
        item = self.create('MeetingItem')
        wf_conditions = item.wfConditions()
        # call _getLastValidationState 1000 times
        self._check__getLastValidationState(wf_conditions, times=1000)

    @timecall
    def _check__getLastValidationState(self, wf_conditions, times=1):
        ''' '''
        pm_logger.info('Call _getLastValidationState {0} times'.format(times))
        for time in range(times):
            wf_conditions._getLastValidationState()

    def test_pm_Speed_get_full_title(self):
        '''Test organization.get_full_title.'''
        self.changeUser('pmManager')
        org = get_organizations()[0]
        # call get_full_title 1000 times
        self._check_get_full_title(org, times=1000)

    @timecall
    def _check_get_full_title(self, org, times=1):
        ''' '''
        pm_logger.info('Call organization.get_full_title {0} times'.format(times))
        for time in range(times):
            org.get_full_title()

    def test_pm_Speed_AsyncLoadItemAssemblyAndSignatures(self):
        """Test when loading @@load_item_assembly_and_signatures
           with 40 assemby members and 25 votes."""
        self.changeUser('siteadmin')
        cfg = self.meetingConfig
        self._removeConfigObjectsFor(cfg)
        # create 40 new held positions
        pm_logger.info('Creating 40 held positions...')
        for i in range(40):
            person = self.create('person')
            self.create('held_position',
                        folder=person,
                        label=str(i),
                        usages=['assemblyMember'],
                        defaults=['present', 'voter'])
        self._setUpOrderedContacts()
        # now create an item with 1 vote and one with 25 votes
        pm_logger.info('Creating meeting, items and votes...')
        self.changeUser('pmManager')
        public_item1 = self.create('MeetingItem')
        public_item25 = self.create('MeetingItem')
        secret_item1 = self.create('MeetingItem', pollType='secret')
        secret_item25 = self.create('MeetingItem', pollType='secret')
        meeting = self.create('Meeting')
        self.presentItem(public_item1)
        self.presentItem(public_item25)
        self.presentItem(secret_item1)
        self.presentItem(secret_item25)
        # setup public votes
        voters = meeting.get_voters()
        public_votes = public_item1.get_item_votes()[0]
        for voter in voters:
            public_votes['voters'][voter] = "yes"
        meeting.set_item_public_vote(public_item1, public_votes, 0)
        for i in range(25):
            meeting.set_item_public_vote(public_item25, public_votes, i)

        # setup secret votes
        secret_votes = secret_item1.get_item_votes()[0]
        secret_votes['votes']['abstain'] = 20
        secret_votes['votes']['no'] = 10
        secret_votes['votes']['yes'] = 5
        meeting.set_item_secret_vote(secret_item1, secret_votes, 0)
        for i in range(25):
            meeting.set_item_secret_vote(secret_item25, secret_votes, i)
        self._check_load_item_assembly_and_signatures(public_item1, True, 1)
        self._check_load_item_assembly_and_signatures(public_item25, True, 25)
        self._check_load_item_assembly_and_signatures(secret_item1, False, 1)
        self._check_load_item_assembly_and_signatures(secret_item25, False, 25)

    @timecall
    def _check_load_item_assembly_and_signatures(self, item, public=True, votes=1):
        ''' '''
        public = "public" if public else "secret"
        pm_logger.info('Call @@load_item_assembly_and_signatures for "{0}" '
                       'item with "{1}" vote(s)'.format(public, votes))
        item.restrictedTraverse('@@load_item_assembly_and_signatures')()

    def test_pm_Speed_MeetingConfigEdit(self):
        """Check editing a MeetingConfig, especially when using
           MeetingConfig.usingGroups as this will update every existing
           meetings and add local roles of selected usingGroups."""
        self.changeUser("siteadmin")
        cfg = self.meetingConfig
        # create 100 groups
        org_uids = []
        pm_logger.info('Creating 100 organizations...')
        for i in range(100):
            org = self.create('organization', id=i, title='Org %d' % i)
            org_uid = org.UID()
            self._select_organization(org_uid)
            org_uids.append(org_uid)
        pm_logger.info('Done.')
        cfg.setUsingGroups(org_uids)
        # create 250 meetings
        pm_logger.info('Creating 100 meetings...')
        for i in range(100):
            self.create('Meeting', date=datetime.now() + timedelta(i + 1))
        pm_logger.info('Done.')
        self._check_meeting_config_edit()

    @timecall
    def _check_meeting_config_edit(self):
        ''' '''
        pm_logger.info('Save MeetingConfig')
        notify(ObjectEditedEvent(self.meetingConfig))


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testPerformances, prefix='test_pm_'))
    return suite



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from imio.helpers.security import fplog
from plone.z3cform.layout import wrap_form
from Products.PloneMeeting.browser.advices import BaseAdviceInfoForm
from Products.PloneMeeting.browser.advices import IBaseAdviceInfoSchema
from Products.PloneMeeting.config import PMMessageFactory as _
from z3c.form import button
from zope import schema
from zope.i18n import translate
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory


def get_item(context):
    """Context may be an item or an advice."""
    if context.__class__.__name__ == "MeetingItem":
        return context
    else:
        return context.aq_parent


@provider(IContextAwareDefaultFactory)
def proposing_group_comment_default(context):
    """
      Get the adviser_id from the REQUEST and get the comment stored
      in MeetingItem.adviceIndex.
    """
    adviser_id = context.REQUEST.get('advice_id', u'')
    item = get_item(context)
    return item.adviceIndex[adviser_id]['proposing_group_comment']


class IAdviceProposingGroupComment(IBaseAdviceInfoSchema):

    proposing_group_comment = schema.Text(
        title=_(u"Comment"),
        description=_(u""),
        defaultFactory=proposing_group_comment_default,
        required=False)


class AdviceProposingGroupCommentForm(BaseAdviceInfoForm):
    """
    """
    label = _(u"Advice proposing group comment")
    description = _(u"Advice proposing group comment description")
    schema = IAdviceProposingGroupComment
    ignoreContext = True  # don't use context to get widget data

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _init(self, data):
        """ """
        self.item = get_item(self.context)
        self.advice_infos = self._advice_infos(data, self.item)
        self.label = u"{1} - {0}".format(
            translate(self.label, context=self.request),
            translate(
                "Advice of ${advice_name}",
                mapping={'advice_name':
                         self.context.adviceIndex[data['advice_uid']]['name']},
                domain='PloneMeeting',
                context=self.request))

    def _check_auth(self, data):
        """Raise Unauthorized if current user may not view or edit comment."""
        # initialize some values on self
        self._init(data)
        if not self.mayEditProposingGroupComment() or \
           not self.mayViewProposingGroupComment():
            raise Unauthorized

    def mayEditProposingGroupComment(self):
        """ """
        return self.advice_infos.mayEditProposingGroupComment()

    def mayViewProposingGroupComment(self):
        """ """
        return self.advice_infos.mayViewProposingGroupComment()

    @button.buttonAndHandler(_('save'), name='Save')
    def handleSave(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        # check that user is not trying to workaround security
        self._check_auth(data)

        # save proposing_group_comment and return, make sure we do not set None
        self.item.adviceIndex[data['advice_uid']]['proposing_group_comment'] = \
            data['proposing_group_comment'] or u""
        # make sure advice cache is invalidated as proposing group comment
        # is displayed on advice view and does not change the advice modified date
        advice = self.item.getAdviceObj(data['advice_uid'])
        # redirect to item or advice view on correct anchor
        self.request.RESPONSE.redirect(
            self.context.absolute_url() + "#adviceAndAnnexes")
        # invalidate advice view cache
        if advice is not None:
            # as we use etags, we will change the _p_mtime because
            # doNotCache(advice, self.request, self.request.RESPONSE)
            # seems to work with FF but not with Chrome...
            # Setting an arbitrary attribute will update _p_mtime
            # description is not used but exists on any DX content
            advice.description = advice.description
        extras = 'item={0} advice_id={1}'.format(
            repr(self.item), data['advice_uid'])
        fplog('edit_proposing_group_comment', extras=extras)

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self.request.RESPONSE.redirect(self.context.absolute_url())

    def update(self):
        """ """
        super(AdviceProposingGroupCommentForm, self).update()
        self._check_auth({"advice_uid": self.widgets['advice_uid'].value})
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')


AdviceProposingGroupCommentFormWrapper = wrap_form(AdviceProposingGroupCommentForm)



# -*- coding: utf-8 -*-

from collective.contact.core.content.person import IPerson
from collective.contact.core.content.person import Person
from plone.autoform import directives as form
from plone.dexterity.schema import DexteritySchemaPolicy
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.events import _invalidateAttendeesRelatedCache
from zope import schema


class IPMPerson(IPerson):
    """ """

    form.order_before(firstname_abbreviated='gender')
    firstname_abbreviated = schema.TextLine(
        title=_("Firstname abbreviated"),
        required=False,
    )

    form.order_before(signature='photo')


class PMPerson(Person):
    """ """

    def get_held_position_by_type(self, position_type=None):
        """Get held_position by type.
           If p_position_type is None, returns first found held_position."""
        held_positions = self.get_held_positions()
        for held_position in held_positions:
            if not position_type or held_position.position_type == position_type:
                return held_position

    def _invalidateCachedMethods(self):
        '''Clean cache for vocabularies using held_positions.'''
        _invalidateAttendeesRelatedCache()


class PMPersonSchemaPolicy(DexteritySchemaPolicy):
    """ """

    def bases(self, schemaName, tree):
        return (IPMPerson, )
